/*
 * FPGA Manager Driver for Altera FPGAs used in Bercut ETN
 *
 *  Copyright (C) 2013 Altera Corporation
 *  Copyright (C) 2015 STC Metrotek 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#define pr_fmt(fmt) "ETN_FPGA: %s:%d: " fmt, __FUNCTION__, __LINE__

#include <linux/kernel.h>
#include <linux/of_device.h>
#include <linux/of_platform.h>
#include <linux/of_address.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/cdev.h>
#include <linux/list.h>
#include <linux/mm.h>
#include <linux/pm.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/completion.h>
#include <linux/fpga.h>
#include <linux/firmware.h>
#include <asm/atomic.h>

#include "main.h"
#include "altera.h"

static int alt_fpga_mon_status_get(struct fpga_manager *mgr)
{
	return fpga_mgr_reg_readl(mgr, ALT_FPGAMGR_MON_GPIO_EXT_PORTA_OFST) &
		ALT_FPGAMGR_MON_STATUS_MASK;
}

int alt_fpga_state_get(struct fpga_manager *mgr)
{
	if ((alt_fpga_mon_status_get(mgr) & ALT_FPGAMGR_MON_FPGA_POWER_ON) == 0)
		return ALT_FPGAMGR_STAT_POWER_OFF;

	return fpga_mgr_reg_readl(mgr, ALT_FPGAMGR_STAT_OFST) &
		ALT_FPGAMGR_STAT_STATE_MASK;
}

/** 
 * Initialize hardware interfaces.
 *
 * 1. Disable reset in RSTMGR's brgmodrst register.
 * 2. Enable remapping in L3REGS's remap register.
 * 3. Unmap RSTMGR and L3REGS memory.
 *
 * @param etn - etn device descriptor
 * @return 0 on success, -ENXIO on error.
 */ 
static int alt_init_hw_ifaces(void)
{
	void __iomem *RSTMGR = NULL;
	void __iomem *L3REGS = NULL;
	uint32_t brgmodrst = 0;
	uint32_t test = 0;
	uint32_t remap = 0;
	int ret = 0;

	// ------------------------------------------
	// First, map configuration registers regions
	// ------------------------------------------
	RSTMGR = ioremap(RSTMGR_ADDR, RSTMGR_LEN);
	if (!RSTMGR) {
		pr_err("Failed to map RSTMGR address space \n");
		ret = -ENXIO;
		goto err_release_rstmgr;
	}

	L3REGS = ioremap(L3REGS_ADDR, L3REGS_LEN);
	if (!L3REGS) {
		pr_err("Failed to map L3REGS ddress space \n");
		ret = -ENXIO;
		goto err_release_l3regs;
	}

	// ---------------------------------------
	// Now do mapping related hw configuration
	// ---------------------------------------

	// 1. Disable reset
	brgmodrst = ~(BRGMODRST_H2F_BIT | BRGMODRST_LWH2F_BIT | BRGMODRST_F2H_BIT) & BRGMODRST_MASK;
	iowrite32(brgmodrst, RSTMGR + BRGMODRST_REG_OFF);

	// Check how it was written
	test = ioread32(RSTMGR + BRGMODRST_REG_OFF);
	if (test != brgmodrst) {
		pr_err("Failed to disable H2F, LWH2F and F2H reset.\n");
		pr_err("brgmodrst is 0x%4X, must be 0x%4X\n", test, brgmodrst);
		ret = -ENXIO;
		goto err_release_l3regs;
	}

	// 2. Enable remaping
	remap = L3REGS_REMAP_LWH2F_BIT | L3REGS_REMAP_H2F_BIT;
	iowrite32(remap, L3REGS + L3REGS_REMAP_OFF);
	// remap register is write-only, so we can't check it.

	// We don't need L3 regs and reset management anymore
err_release_l3regs:
	if (L3REGS)
		iounmap(L3REGS);

err_release_rstmgr:
	if (RSTMGR)
		iounmap(RSTMGR);

	return ret;
}


int alt_fpga_fpga2sdram_cfg(int32_t val)
{
	void __iomem *FPGAPORTRST = NULL;

	FPGAPORTRST = ioremap(FPGAPORTRST_ADDR, FPGAPORTRST_LEN);
	if (!FPGAPORTRST) {
		pr_err("Failed to map FPGAPORTRST to disable fpga2sdram interface\n");
		return -ENXIO;
	}

	/* Disable fpga2sdram interface */
	iowrite32(val, FPGAPORTRST);

	iounmap(FPGAPORTRST);
	return 0;
}

/*
 * Set the DCLKCNT, wait for DCLKSTAT to report the count completed, and clear
 * the complete status.
 */
static int alt_fpga_dclk_set_and_wait_clear(struct fpga_manager *mgr, u32 count)
{
	int timeout = 2;
	u32 done;

	/* Clear any existing DONE status. */
	if (fpga_mgr_reg_readl(mgr, ALT_FPGAMGR_DCLKSTAT_OFST))
		fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_DCLKSTAT_OFST,
			      ALT_FPGAMGR_DCLKSTAT_DCNTDONE_E_DONE);

	/* Issue the DCLK count. */
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_DCLKCNT_OFST, count);

	/* Poll DCLKSTAT to see if it completed in the timeout period. */
	do {
		done = fpga_mgr_reg_readl(mgr, ALT_FPGAMGR_DCLKSTAT_OFST);
		if (done == ALT_FPGAMGR_DCLKSTAT_DCNTDONE_E_DONE) {
			/* clear the DONE status. */
			fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_DCLKSTAT_OFST,
				      ALT_FPGAMGR_DCLKSTAT_DCNTDONE_E_DONE);
			return 0;
		}
		if (count <= 4)
			udelay(1);
		else
			// Was msleep(20), but didn't work for CONFIG_PREEMPT 
			// (Sleeping in atomic context).
			// This way it works for both preemptive and non-preemptive kernels.
			mdelay(20); 

	} while (timeout--);

	return -ETIMEDOUT;
}

static int alt_fpga_wait_for_state(struct fpga_manager *mgr, u32 state)
{
	int timeout = 2;

	/*
	 * HW doesn't support an interrupt for changes in state, so poll to see
	 * if it matches the requested state within the timeout period.
	 */
	do {
		if ((alt_fpga_state_get(mgr) & state) != 0)
			return 0;

		// Was msleep(20), but didn't work for CONFIG_PREEMPT
		// (Sleeping in atomic context).
		// This way it works for both preemptive and non-preemptive kernels.
		mdelay(20);
	} while (timeout--);

	return -ETIMEDOUT;
}

static void alt_fpga_enable_irqs(struct fpga_manager *mgr, u32 irqs)
{
	/* set irqs to level sensitive */
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_MON_GPIO_INTTYPE_LEVEL_OFST, 0);

	/* set interrupt polarity */
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_MON_GPIO_INT_POL_OFST, irqs);

	/* clear irqs */
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_MON_GPIO_PORTA_EOI_OFST, irqs);

	/* unmask interrupts */
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_MON_GPIO_INTMSK_OFST, 0);

	/* enable interrupts */
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_MON_GPIO_INTEN_OFST, irqs);
}

static void alt_fpga_disable_irqs(struct fpga_manager *mgr)
{
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_MON_GPIO_INTEN_OFST, 0);
}

static irqreturn_t alt_fpga_isr(int irq, void *dev_id)
{
	struct fpga_manager *mgr = dev_id;
	u32 irqs, st;
	bool conf_done, nstatus;

	/* clear irqs */
	irqs = fpga_mgr_reg_raw_readl(mgr, ALT_FPGAMGR_MON_GPIO_INTSTAT_OFST);
	fpga_mgr_reg_raw_writel(mgr, ALT_FPGAMGR_MON_GPIO_PORTA_EOI_OFST, irqs);

	st = fpga_mgr_reg_raw_readl(mgr, ALT_FPGAMGR_MON_GPIO_EXT_PORTA_OFST);
	conf_done = (st & ALT_FPGAMGR_MON_CONF_DONE) != 0;
	nstatus = (st & ALT_FPGAMGR_MON_NSTATUS) != 0;

	/* success */
	if (conf_done && nstatus) {
		/* disable irqs */
		fpga_mgr_reg_raw_writel(mgr, ALT_FPGAMGR_MON_GPIO_INTEN_OFST, 0);
		complete(&mgr->status_complete);
	}

	return IRQ_HANDLED;
}

static int alt_fpga_wait_for_config_done(struct fpga_manager *mgr)
{
	int timeout, ret = 0;

	alt_fpga_disable_irqs(mgr);
	reinit_completion(&mgr->status_complete);
	alt_fpga_enable_irqs(mgr, ALT_FPGAMGR_MON_CONF_DONE);

	timeout = wait_for_completion_interruptible_timeout(
						&mgr->status_complete,
						msecs_to_jiffies(10));
	if (timeout == 0) {
		dev_err(mgr->parent, "timeout\n");
		ret = -ETIMEDOUT;
	}

	alt_fpga_disable_irqs(mgr);
	return ret;
}

/* For ALT_FPGAMGR_STAT_MSEL field */
static struct cfgmgr_mode cfgmgr_modes[] = {
	[MSEL_PP16_FAST_NOAES_NODC] = { CFGWDTH_16 | CDRATIO_X1, 1 },
	[MSEL_PP16_FAST_AES_NODC] =   { CFGWDTH_16 | CDRATIO_X2, 1 },
	[MSEL_PP16_FAST_AESOPT_DC] =  { CFGWDTH_16 | CDRATIO_X4, 1 },
	[MSEL_PP16_SLOW_NOAES_NODC] = { CFGWDTH_16 | CDRATIO_X1, 1 },
	[MSEL_PP16_SLOW_AES_NODC] =   { CFGWDTH_16 | CDRATIO_X2, 1 },
	[MSEL_PP16_SLOW_AESOPT_DC] =  { CFGWDTH_16 | CDRATIO_X4, 1 },
	[MSEL_PP32_FAST_NOAES_NODC] = { CFGWDTH_32 | CDRATIO_X1, 1 },
	[MSEL_PP32_FAST_AES_NODC] =   { CFGWDTH_32 | CDRATIO_X4, 1 },
	[MSEL_PP32_FAST_AESOPT_DC] =  { CFGWDTH_32 | CDRATIO_X8, 1 },
	[MSEL_PP32_SLOW_NOAES_NODC] = { CFGWDTH_32 | CDRATIO_X1, 1 },
	[MSEL_PP32_SLOW_AES_NODC] =   { CFGWDTH_32 | CDRATIO_X4, 1 },
	[MSEL_PP32_SLOW_AESOPT_DC] =  { CFGWDTH_32 | CDRATIO_X8, 1 },
};

static int alt_fpga_cfg_mode_get(struct fpga_manager *mgr)
{
	u32 msel;

	msel = fpga_mgr_reg_readl(mgr, ALT_FPGAMGR_STAT_OFST);
	msel &= ALT_FPGAMGR_STAT_MSEL_MASK;
	msel >>= ALT_FPGAMGR_STAT_MSEL_SHIFT;

	/* Check that this MSEL setting is supported */
	if ((msel >= sizeof(cfgmgr_modes)/sizeof(struct cfgmgr_mode)) ||
	     !cfgmgr_modes[msel].valid) {
		dev_warn(mgr->parent, "Invalid MSEL setting");
		return -1;
	}

	return msel;
}

static int alt_fpga_cfg_mode_set(struct fpga_manager *mgr)
{
	u32 ctrl_reg, mode;

	/* get value from MSEL pins */
	mode = alt_fpga_cfg_mode_get(mgr);
	if (mode < 0)
		return -EINVAL;

	/* Adjust CTRL for the CDRATIO */
	ctrl_reg = fpga_mgr_reg_readl(mgr, ALT_FPGAMGR_CTL_OFST);
	ctrl_reg &= ~ALT_FPGAMGR_CTL_CDRATIO_MASK;
	ctrl_reg &= ~ALT_FPGAMGR_CTL_CFGWDTH_MASK;
	ctrl_reg |= cfgmgr_modes[mode].ctrl;

	/* Set NCE to 0. */
	ctrl_reg &= ~ALT_FPGAMGR_CTL_NCE;
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_CTL_OFST, ctrl_reg);

	return 0;
}

static int alt_fpga_reset(struct fpga_manager *mgr)
{
	u32 ctrl_reg, status;

	/*
	 * Step 3: Set CTRL.NCONFIGPULL to 1 to put FPGA in reset
	 */
	ctrl_reg = fpga_mgr_reg_readl(mgr, ALT_FPGAMGR_CTL_OFST);
	ctrl_reg |= ALT_FPGAMGR_CTL_NCFGPULL;
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_CTL_OFST, ctrl_reg);

	/*
	 * Step 4: Wait for STATUS.MODE to report FPGA is in reset phase
	 */
	status = alt_fpga_wait_for_state(mgr, ALT_FPGAMGR_STAT_RESET);

	/*
	 * Step 5: Set CONTROL.NCONFIGPULL to 0 to release FPGA from reset
	 */
	ctrl_reg &= ~ALT_FPGAMGR_CTL_NCFGPULL;
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_CTL_OFST, ctrl_reg);

	if (status) {
		/* This is a failure from Step 4. */
		dev_err(mgr->parent,
			"Error in step 4: Wait for RESET timeout.\n");
		return -ETIMEDOUT;
	}

	return 0;
}

/*
 * Prepare the FPGA to receive the configuration data.
 */
static int alt_fpga_configure_init(struct fpga_manager *mgr)
{
	int ret;
	struct etn_fpga *etn_fpga = mgr->priv;

	if (atomic_read(&etn_fpga->refcnt) != 0) {
		/* 
		 * IMPORTANT NOTE:
		 * FPGA_MGR_DEV_BUSY flag set in fpga-mgr-core 'open' method and unset
		 * in 'release'. But when we return error here, 'open' is failed and
		 * 'release' is not invoked and flag is set forever. Thus we clear it
		 * manually here before returning error.
		 */
		clear_bit_unlock(FPGA_MGR_DEV_BUSY, &mgr->flags);
		pr_warn("Lock contention\n");
		return -EBUSY;
	}

	/* Forget old features */
	fpgafeat_deinit(&etn_fpga->features);

	/* Fix fpga2sdram WAITREQUEST signal hangup.
	 * Before flashing FPGA we disable fpga2sdram interface,
	 * and in the end (alt_fpga_configure_complete) we enable it back.
	 */
	if (alt_fpga_fpga2sdram_cfg(FPGAPORTRST_DISABLE_ALL)) {
		dev_err(mgr->parent, "Error while disabling fpga2sdram interface\n");
		return -EIO;
	}

	/*
	 * Step 1:
	 *  - Set CTRL.CFGWDTH, CTRL.CDRATIO to match cfg mode
	 *  - Set CTRL.NCE to 0
	 */
	ret = alt_fpga_cfg_mode_set(mgr);
	if (ret) {
		pr_warn("alt_fpga_cfg_mode_get ret is %d\n", ret);
		return ret;
	}

	/* Step 2: Set CTRL.EN to 1 */
	fpga_mgr_reg_set_bitsl(mgr, ALT_FPGAMGR_CTL_OFST, ALT_FPGAMGR_CTL_EN);

	/* Steps 3 - 5: Reset the FPGA */
	ret = alt_fpga_reset(mgr);
	if (ret) {
		pr_warn("alt_fpga_reset ret is %d\n", ret);
		return ret;
	}

	/* Step 6: Wait for FPGA to enter configuration phase */
	if (alt_fpga_wait_for_state(mgr, ALT_FPGAMGR_STAT_CFG)) {
		dev_err(mgr->parent,
			"Error in step 6: Wait for CFG timeout.\n");
		return -ETIMEDOUT;
	}

	/* Step 7: Clear nSTATUS interrupt */
	fpga_mgr_reg_writel(mgr, ALT_FPGAMGR_MON_GPIO_PORTA_EOI_OFST,
			    ALT_FPGAMGR_MON_NSTATUS);

	/* Step 8: Set CTRL.AXICFGEN to 1 to enable transfer of config data */
	fpga_mgr_reg_set_bitsl(mgr, ALT_FPGAMGR_CTL_OFST,
			       ALT_FPGAMGR_CTL_AXICFGEN);

	return 0;
}

/*
 * Step 9: write data to the FPGA data register
 */
static ssize_t alt_fpga_configure_write(struct fpga_manager *mgr,
					char *buf, size_t count)
{
	u32 *buffer_32 = (u32 *)buf;
	size_t total = count;
	size_t i = 0;

	if (count <= 0)
		return 0;

	pr_debug("Writing %zu bytes\n", count);

	/* Write out the complete 32-bit chunks. */
	while (count >= sizeof(u32)) {
		fpga_mgr_data_writel(mgr, buffer_32[i++]);
		count -= sizeof(u32);
	}

	/* Write out remaining non 32-bit chunks. */
	switch (count) {
	case 3:
		fpga_mgr_data_writel(mgr, buffer_32[i++] & 0x00ffffff);
		break;
	case 2:
		fpga_mgr_data_writel(mgr, buffer_32[i++] & 0x0000ffff);
		break;
	case 1:
		fpga_mgr_data_writel(mgr, buffer_32[i++] & 0x000000ff);
		break;
	default:
		/* This will never happen. */
		break;
	}

	return total;
}

static int alt_fpga_configure_complete(struct fpga_manager *mgr)
{
	u32 status = 0;
	struct fpga_feature_list *flist = NULL;
	struct etn_fpga *etn_fpga = mgr->priv;

	if (!etn_fpga) {
		dev_err(mgr->parent, "Invalid etn_fpga struct. Can't finish.\n");
		return -EINVAL;
	}

	pr_debug("Finishing FPGA flashing\n");

	/*
	 * Step 10:
	 *  - Observe CONF_DONE and nSTATUS (active low)
	 *  - if CONF_DONE = 1 and nSTATUS = 1, configuration was successful
	 *  - if CONF_DONE = 0 and nSTATUS = 0, configuration failed
	 */
	status = alt_fpga_wait_for_config_done(mgr);
	if (status) {
		dev_err(mgr->parent, "FPGA configuration is not done\n");
		return -EIO;
	}

	/* Step 11: Clear CTRL.AXICFGEN to disable transfer of config data */
	fpga_mgr_reg_clr_bitsl(mgr, ALT_FPGAMGR_CTL_OFST,
			       ALT_FPGAMGR_CTL_AXICFGEN);

	/*
	 * Step 12:
	 *  - Write 4 to DCLKCNT
	 *  - Wait for STATUS.DCNTDONE = 1
	 *  - Clear W1C bit in STATUS.DCNTDONE
	 */
	if (alt_fpga_dclk_set_and_wait_clear(mgr, 4)) {
		dev_err(mgr->parent, "Error: Wait for dclk(4) timeout.\n");
		return -ETIMEDOUT;
	}

#if _ALT_FPGA_USE_DCLK
	/* Step 13: Wait for STATUS.MODE to report INIT or USER MODE */
	if (alt_fpga_wait_for_state(mgr, ALT_FPGAMGR_STAT_INIT |
					 ALT_FPGAMGR_STAT_USER_MODE)) {
		dev_err(mgr->parent,
			"Error in step 13: Wait for USER_MODE timeout.\n");
		return -ETIMEDOUT;
	}

	/*
	 * Extra steps for Configuration with DCLK for Initialization Phase
	 * Step 14 (using 4.2.1.2 steps), 15 (using 4.2.1.2 steps)
	 *  - Write 0x5000 to DCLKCNT == the number of clocks needed to exit
	 *    the Initialization Phase.
	 *  - Poll until STATUS.DCNTDONE = 1, write 1 to clear
	 */
	if (alt_fpga_dclk_set_and_wait_clear(mgr, 0x5000)) {
		dev_err(mgr->parent,
			"Error in step 15: Wait for dclk(0x5000) timeout.\n");
		return -ETIMEDOUT;
	}
#endif

	/* Step 13: Wait for STATUS.MODE to report USER MODE */
	if (alt_fpga_wait_for_state(mgr, ALT_FPGAMGR_STAT_USER_MODE)) {
		dev_err(mgr->parent,
			"Error in step 13: Wait for USER_MODE timeout.\n");
		return -ETIMEDOUT;
	}

	/* Step 14: Set CTRL.EN to 0 */
	fpga_mgr_reg_clr_bitsl(mgr, ALT_FPGAMGR_CTL_OFST, ALT_FPGAMGR_CTL_EN);


	if (alt_fpga_fpga2sdram_cfg(FPGAPORTRST_ENABLE_ALL)) {
		dev_err(mgr->parent, "Error in enabling fpga2sdram interface\n");
		return -EIO;
	}

	if (alt_init_hw_ifaces()) {
		dev_err(mgr->parent, "Failed to unreset FPGA\n");
		return -EIO;
	}

	/* Read FPGA features from new firmware */
	flist = fpgafeat_init(&etn_fpga->pdev->dev);
	if (!flist) {
		dev_err(mgr->parent, "Failed to initialize FPGA features\n");
		return -EIO;
	}
	etn_fpga->features = flist;
	
	return 0;
}

static int alt_fpga_status(struct fpga_manager *mgr, char *buf)
{
	u32 state;
	const char *state_str = NULL;
	int ret;

	state = alt_fpga_state_get(mgr);

	if (state < sizeof(altera_fpga_state_str)/sizeof(char *))
		state_str = altera_fpga_state_str[state];

	if (state_str)
		ret = sprintf(buf, "%s\n", state_str);
	else
		ret = sprintf(buf, "%s\n", "unknown state");

	return ret;
}

struct fpga_manager_ops altera_fpga_mgr_ops = {
	.status = alt_fpga_status,
	.write_init = alt_fpga_configure_init,
	.write = alt_fpga_configure_write,
	.write_complete = alt_fpga_configure_complete,
	.isr = alt_fpga_isr,
};


/**
 * Load bytes array fw_data of size fw_size into FPGA that is managed by mgr.
 *
 * This function mimics FPGA reflash from userspace command 
 * 
 *     cat fpga.rbf > /dev/fpga0
 * 
 * You should NOT call it whenever you want because you don't have any
 * concurrency control. The only case it's used is from etn_fpga_probe, i.e.
 * when module is loading. When it's called, char device is not exported yet,
 * hence it's *guaranteed* that nobody will concurrently access FPGA while in
 * this function.
 * 
 * @param fw_data - Array of bytes with FPGA firmware
 * @param fw_size - Size of FPGA firmware array
 * @param mgr     - Pointer to FPGA manager
 *
 * @return 0 on success, error code otherwise
 */
int alt_fpga_load_fw(char *fw_data, size_t fw_size, struct fpga_manager *mgr)
{
	int rc;
	ssize_t written;
	
	rc = alt_fpga_configure_init(mgr);
	if (rc) {
		pr_err("Failed to prepare FPGA for flashing\n");
		return rc;
	}

	written = alt_fpga_configure_write(mgr, fw_data, fw_size);
	if (written != fw_size) {
		pr_err("Failed to write firmware to FPGA\n");
		return -EIO;
	}

	rc = alt_fpga_configure_complete(mgr);
	if (rc) {
		pr_err("Failed to complete FPGA flashing\n");
		return rc;
	}

	return 0;
}

