#define pr_fmt(fmt) "ETN_FPGA: %s:%d: " fmt, __FUNCTION__, __LINE__

#include <linux/slab.h>
#include <linux/ioport.h>
#include <asm/io.h>

#include "fpgafeat.h"
#include "features.h"
#include "features_sysfs.h"

/** Size of memory that stores memory information, 4-byte words */
#define FEATURE_MEM_WORDS (256)

/* Length of feature memory word, bytes */
#define FEATURE_MEM_WORD_LEN (sizeof(uint32_t))

static signed int opt_lookup(const char *what, const char** where) 
{ 
	unsigned short i = 0; 

	/* Short, but dangerous code */ 
	for (i = 0; where[i] != NULL;  i ++) { 
		if (strcmp(where[i], what) == 0)
			return i; 
	} 
	return -1; 
}

/** 
 * Add a feature to the struct fpga_feature. 
 *
 * Function reallocates memory of features list, so user should call
 * free_features_list function to avoid memory leaks.
 *
 * \param list		 Pointer to features list.
 * \param feature Pointer to feature for adding.
 *
 * \return 0 if feature adding is successful, negative integer otherwise.
 */
int add_feature_to_list(struct fpga_feature_list *list, struct fpga_feature *feature)
{
	void *tmp;
	void *ptr = (void *)list->features_arr;

	struct fpga_feature *feat = get_feature_by_name(feature->name, list);

	if (feat == NULL) {
		tmp = krealloc(ptr, sizeof(*feature) * (list->count + 1), GFP_KERNEL);
		if (tmp == NULL) {
			pr_err("Can't allocate memory to add feature\n");
			return -1; 
		}
		ptr = tmp;

		list->features_arr = (struct fpga_feature *)ptr;
		memcpy(&list->features_arr[list->count], feature, sizeof(*feature));
		list->count++;
	} else {
		memcpy(feat, feature, sizeof(*feature));
	}

	return 0;
}

/** 
 * Function frees memory allocated to feature list.
 *
 * \param list Pointer to struct fpga_feature variable to free.
 */
void free_features_list(struct fpga_feature_list *list)
{
	list->count = 0;
	if (list->features_arr != NULL) {
		kfree(list->features_arr);
		list->features_arr = NULL;
	}
}

/** 
 * Function reads a 4-byte word from memory contains feature info.
 *
 * \param addr Memory byte address to read from.
 * \param data Pointer to store read data.
 *
 * \return 0 if there are no errors, negative integer otherwise.
 */
static int read_features_memory_word(void __iomem *base, uint16_t word, uint32_t *data)
{
	uint16_t tmp0, tmp1;

/* Control register with features */
#define FPGA_ID_CR 2

	/* FPGA features lies in memory with indirect address mode.
	 * To access it, we store word address in FPGA_ID_CR on H2F bus.
	 * And then read data from first two registers.*/
	iowrite16(word & 0xFF, base + FPGA_ID_CR * sizeof(word));

	tmp0 = ioread16(base);
	tmp1 = ioread16(base + (1 * sizeof(word)));

	*data |= tmp0;
	*data |= tmp1 << 16;

	return 0;
}

/** 
 * Read features from FPGA.
 *
 * \param buf Pointer to buffer to store read date. It must have size at least
 * FEATURE_MEM_WORDS 4-bytes word. 
 *
 * \return 0 if there are no errors, negative integer otherwise.
 */
static int read_features(uint32_t *buffer)
{
	uint16_t i;
	int err = 0;
	void __iomem *base;

#define   H2F_ADDR     0xC0000000
#define   H2F_LEN      0x4088

	base = ioremap(H2F_ADDR, H2F_LEN);
	if (!base) {
		pr_err("Failed to map H2F address space\n");
		return -1;
	}

	for (i = 0; i < FEATURE_MEM_WORDS; i++) {
		err = read_features_memory_word(base, i, &buffer[i]);
		if (err) {
			pr_err("Can't read features info from FPGA\n");
			goto exit;
		}
	}

exit:
	iounmap(base);
	return err;
}

/** 
 * Function parses an FPGA feature and form struct fpga_feature
 * structure that describes it.
 *
 * \param buffer  Input buffer to parse.
 * \param idx		 Feature index.
 * \param feature Pointer to struct fpga_feature variable to store parsed data.
 *
 * \return 0 if there are no errors, negative integer otherwise.
 */
static int parse_feature(const uint32_t *buffer, unsigned int idx, 
                         struct fpga_feature *feature)
{
	int i;
	char *tmp;
	int name_idx;
	const uint32_t *ptr = buffer;
	uint8_t individual_bit = *ptr & 0x1;
	uint8_t feat_width		 = (*ptr >> 24) & 0xff;
	ptr++;

	pr_debug("Feature %d: I_bit %d, length %d\n", idx, individual_bit, feat_width);

	feature->type = idx;

	tmp = (char *)ptr;

	for (i = 0; i < FEATURE_NAME_MAX; i++) {
		feature->name[i] = tmp[(FEATURE_NAME_MAX - 1) - i];
	}
	feature->name[FEATURE_NAME_MAX] = '\0';
	ptr += FEATURE_NAME_MAX / FEATURE_MEM_WORD_LEN;

	pr_debug("Feature %d: name %s\n", idx, feature->name);

	name_idx = opt_lookup(feature->name, fpga_feature_names);
	if (name_idx >= 0) {
		strncpy(feature->description, fpga_feature_descriptors[name_idx], 
				sizeof(feature->description) - 1);
	} else {
		strncpy(feature->description, "Description unvaliable\n", 
				sizeof(feature->description) - 1);
	}

	feature->sr_base = *ptr & 0xffff;
	feature->cr_base = (*ptr >> 16) & 0xffff;
	ptr++;

	feature->sr_cnt = *ptr & 0xffff;
	feature->cr_cnt = (*ptr >> 16) & 0xffff;
	ptr++;

	pr_debug("Feature %d: cr_base %d, cr_cnt %d\n", idx, 
			feature->cr_base, feature->cr_cnt);
	pr_debug("Feature %d: sr_base %d, sr_cnt %d\n", idx,
			feature->sr_base, feature->sr_cnt);

	if (individual_bit) {
		feature->port_mask = *ptr;
		ptr++;
	}

	pr_debug("Feature %d: port mask 0x%08x\n", idx, feature->port_mask); 

	return 0;
}

/** 
 * Function calculates offset to feature descriptor for a feature.
 *
 * \param buffer Pointer to memory buffer. It doesn't point to whole memory 
 * start but to feature offset base address. 
 *							 
 * \param idx Feature index.
 *
 * \return offset address from in words from the beginning of ROM memory.
 */
static uint16_t get_feat_romid_addr(const uint32_t *buffer, unsigned int idx)
{
	uint16_t addr;
	unsigned int offset = idx / 2;
	if (idx % 2) {
		addr = (buffer[offset] >> 16) & 0xffff;
	} else {
		addr = buffer[offset] & 0xffff;
	}

	pr_debug("Feature %d address 0x%04x\n", idx, addr);

	return addr;
}

/** 
 * Parse FPGA features from FPGA memory buffer into struct fpga_feature_list.
 * 
 * \param buffer Input buffer to parse.
 * \param list   Pointer to fpga_feature_list struct to store parsed data.
 *
 * \return 0 if there are no errors, negative integer otherwise.
 */
static int parse_features(const uint32_t *buffer, struct fpga_feature_list *list)
{
	int err;
	uint8_t hdr_len;
	unsigned int feature_idx;
	uint64_t feat_bit_mask;
	uint16_t feat_romid_addr;
	const uint32_t *feat_base;
	const uint32_t *ptr = buffer;
	struct fpga_feature feature;

	list->fpga_version = *ptr & 0xffff;
	list->rom_id       = (*ptr >> 16) & 0xff;
	hdr_len	           = (*ptr >> 24) & 0xff;
	ptr++;

	pr_info("FPGA version: 0x%04x\n", list->fpga_version);
	pr_info("ROM ID version: 0x%04x\n", list->rom_id);
	pr_info("Header length: %d\n", hdr_len);

	feat_bit_mask = ((uint64_t)*(ptr + 1) << 32) | *ptr;
	pr_debug("Features bitmask 0x%08llx\n", feat_bit_mask);
	ptr += 2;

	err = 0;
	feature_idx = 0;
	while (feat_bit_mask) {
		if ((feat_bit_mask & 0x1)) {
			feat_romid_addr = get_feat_romid_addr(ptr, feature_idx);
			feat_base = &buffer[feat_romid_addr];

			memset(&feature, 0, sizeof(feature));

			err = parse_feature(feat_base, feature_idx, &feature);
			if (err) {
				pr_err("Error during feature %d parsing\n", feature_idx);
				return -1;
			}

			err = add_feature_to_list(list, &feature);
			if (err) {
				pr_err("Error during adding feature %s to list\n", feature.name);
				return -1;
			}
		}

		feat_bit_mask >>= 1;
		feature_idx++;
	}

	return 0;
}

struct fpga_feature *get_feature( int type, struct fpga_feature_list *list )
{
	int i;
	for (i = 0; i < list->count; i++) {
		if (type == list->features_arr[i].type) {
			return &list->features_arr[i];
		}
	}

	return NULL;
}

/** 
 * Get FPGA feature by name from features list if feature is present.
 *
 * \param name Feature name to get.
 * \param list Features list to search.
 *
 * \return Pointer to struct fpga_feature in case or success, 
 *         NULL if feature is not found.
 */
struct fpga_feature *get_feature_by_name(char *name, struct fpga_feature_list *list)
{
	int i;
	for (i = 0; i < list->count; i++) {
		if (strcmp(name, list->features_arr[i].name) == 0) {
			return &list->features_arr[i];
		}
	}

	return NULL;
}

/** 
 * Return Control Registers base for given port.
 *
 * \param feat Feature to get base, it must be supported on the port.
 * \param port Port index, from 0.
 *
 * \return base address for the port. 
 */
uint16_t get_cr_base_on_port(struct fpga_feature *feat, unsigned int port)
{
	return feat->cr_base + feat->cr_cnt * port;
}
EXPORT_SYMBOL(get_cr_base_on_port);

/** 
 * Function returns Status Registers base for given port.
 *
 *	\param feat Feature to get base, it must be supported on the port.
 *	\param port Port index, from 0.
 *
 *	\return base address for the port. 
 */
uint16_t get_sr_base_on_port(struct fpga_feature *feat, unsigned int port)
{
	return feat->sr_base + feat->sr_cnt * port;
}

struct fpga_feature *get_feature_on_port( int type, unsigned int port, 
                                          struct fpga_feature_list *list )
{
	struct fpga_feature *feat = get_feature(type, list);

	if (feat == NULL) {
		return NULL;
	} 

	if ((feat->port_mask & (1L << port)) == 0) {
		return NULL;
	}

	return feat;
}

/** Function gets FPGA feature from features list if feature is presents
 *				 in the list on the certain port.
 *
 *	\note Control and status offsets for the returned feature are for zero
 *				port.
 *
 *	\param name Feature name to get.
 *	\param port Port number to check, from 0.
 *	\param list Features list to search.
 *
 *	\sa get_cr_base_on_port
 *	\sa get_sr_base_on_port
 *
 *	\return Pointer to struct fpga_feature in case or success, NULL if feature 
 *					is not found.
 */
struct fpga_feature *get_feature_on_port_by_name(char *name, unsigned int port, 
		struct fpga_feature_list *list)
{
	struct fpga_feature *feat = get_feature_by_name(name, list);

	if (feat == NULL) {
		return NULL;
	} 

	if ((feat->port_mask & (1L << port)) == 0) {
		return NULL;
	}

	return feat;
}

/** Function check if FPGA feature is presents in the features list.
 *
 *	\param name Feature name to get.
 *	\param list Features list to search.
 *
 *	\return 1 if feature is present, 0 if it is not.
 */
int is_fpga_feature_present(char *name, struct fpga_feature_list *list)
{
	return (get_feature_by_name(name, list) != NULL) ? 1 : 0;
}

/** Function check if FPGA feature is presents in the features list on
 *				 the certain port
 *
 *	\param name Feature name to get.
 *	\param port Port number to check, from 0.
 *	\param list Features list to search.
 *
 *	\return 1 if feature is present, 0 if it is not.
 */
int is_fpga_feature_present_on_port(char *name, unsigned int port, 
		struct fpga_feature_list *list)
{
	return (get_feature_on_port_by_name(name, port, list) != NULL) ? 1 : 0;
}

/** 
 * Read features data from FPGA, parse it and return pointer to features list.
 */
struct fpga_feature_list *fpgafeat_init(struct device *dev)
{
	struct fpga_feature_list *flist = NULL;
	uint32_t *mem_buffer = NULL;
	int err = 0;

	mem_buffer = kzalloc(FEATURE_MEM_WORDS * sizeof(uint32_t), GFP_KERNEL);
	if (mem_buffer == NULL) {
		pr_err("Failed to allocate memory\n");
		return NULL;
	}

	if (read_features(mem_buffer)) {
		err = -1;
		pr_err("Error during reading FPGA features from memory\n");
		goto fpgafeat_init_err_l;
	}
	pr_debug("Features read from FPGA\n");

	flist = kzalloc(sizeof(*flist), GFP_KERNEL);
	if (!flist) {
		err = -2;
		pr_err("Failed to allocate memory for FPGA features list\n");
		goto fpgafeat_init_err_l;
	}

	if (parse_features(mem_buffer, flist)) {
		err = -3;
		pr_err("Error during parsing FPGA features\n");
		goto fpgafeat_init_err_l;
	}
	pr_debug("Features parsed %p\n", flist);

	if (init_features_sysfs(dev, flist)) {
		err = -4;
		pr_err("Failed to create sysfs hierarchy for FPGA features\n");
		goto fpgafeat_init_err_l;
	}

	return flist;

fpgafeat_init_err_l:

	if (mem_buffer)
		kfree(mem_buffer);

	if (flist) {
		deinit_features_sysfs(flist);
		kfree(flist);
	}

	return NULL;
}

/** Library deinitializer. It clears all memory that has been allocated 
 *				 during library working.
 */
void fpgafeat_deinit(struct fpga_feature_list **plist)
{
	struct fpga_feature_list *list = *plist;
	if (list) {
		deinit_features_sysfs(list);
		free_features_list(list);
		*plist = NULL;
	}
}
