#ifndef ETN_FPGA_MGR_FPGA_FEAT__H
#define ETN_FPGA_MGR_FPGA_FEAT__H

#include <linux/types.h>
#include <linux/device.h>
#include <linux/kobject.h>

/* Maximum length of feature's name in bytes, same as in FPGA memory. */
#define FEATURE_NAME_MAX (16)

/* Maximum length of feature's description, bytes. */
#define FEATURE_DESCRIPTION_MAX (128) 

/** Enum describes all supported FPGA features */
enum fpga_features {
	FPGA_FEATURE_MX_FPGA_ID,
	FPGA_FEATURE_MX_PHY,
	FPGA_FEATURE_MX_STAT,
	FPGA_FEATURE_FLEX_FILT,
	FPGA_FEATURE_FLEX_FILT_PM,
	FPGA_FEATURE_FLEX_FILT_RTP,
	FPGA_FEATURE_FS_FILT,
	FPGA_FEATURE_FIELDS_RPL,
	FPGA_FEATURE_HTP_PACKER,
	FPGA_FEATURE_PKT_ENCAPS,
	FPGA_FEATURE_FLOW_GEN,
	FPGA_FEATURE_SIGN_ANLZ,
	FPGA_FEATURE_NIC_CPU_BUF,
	FPGA_FEATURE_RMON_STAT,
	FPGA_FEATURE_LB,
	FPGA_FEATURE_TWAMP_RX_FILT,
	FPGA_FEATURE_TWAMP_REFL,
	FPGA_FEATURE_NIC_CTRL,
	FPGA_FEATURE_BERT_GEN,
	FPGA_FEATURE_BERT_ANLZ,
	FPGA_FEATURE_LCD,
	FPGA_FEATURE_BEEPER,
        FPGA_FEATURE_SOM_HWTEST,
        FPGA_FEATURE_MPT_ANLZ,
	FPGA_FEATURE_MPT_TX,
	FPGA_FEATURE_FAN, 
	FPGA_FEATURE_PTP_RTC,
};

struct fpga_feature {
	/* To present feature in sysfs */
	struct kobject kobj;
	
	/* Feature type as in enum fpga_features */
	int type;

	/** Feature name. */
	char name[ FEATURE_NAME_MAX + 1 ];

	/** Short description of the feature. */
	char description[ FEATURE_DESCRIPTION_MAX + 1 ];

	/** Mask that shows which ports are supported the feature. '1' means feature
	 *	is supported on corresponding port. */ 
	uint32_t port_mask;

	/** Control registers base address for least port. */
	uint16_t cr_base;
	
	/** Control registers count for each port. */
	uint16_t cr_cnt;

	/** Status registers base address for least port. */
	uint16_t sr_base;

	/** Status registers count for each port. */
	uint16_t sr_cnt;
};
#define to_fpga_feature(x) container_of(x, struct fpga_feature, kobj)

struct fpga_feature_list {

	/* To organize feature in sysfs hierarchy */
	struct kset *kset;

	/** Version of FPGA firmware */
	uint16_t fpga_version;

	/** ROM memory ID version */
	uint8_t rom_id;

	/** Count of features that are supported in the FPGA firmware. */
	int count;

	/** Array of current features. */
	struct fpga_feature *features_arr;
};

struct fpga_feature_list *get_features_list( void );

struct fpga_feature *get_feature( int type, struct fpga_feature_list *list );
struct fpga_feature *get_feature_by_name( char *name, struct fpga_feature_list *list );

struct fpga_feature *get_feature_on_port( int type, unsigned int port, 
		struct fpga_feature_list *list );
struct fpga_feature *get_feature_on_port_by_name( char *name, unsigned int port, 
		struct fpga_feature_list *list );

uint16_t get_cr_base_on_port( struct fpga_feature *feat, unsigned int port );
uint16_t get_sr_base_on_port( struct fpga_feature *feat, unsigned int port );

int is_fpga_feature_present( char *name, struct fpga_feature_list *list );
int is_fpga_feature_present_on_port( char *name, unsigned int port, 
		struct fpga_feature_list *list );

int add_feature_to_list( struct fpga_feature_list *list, struct fpga_feature *feature );
void free_features_list( struct fpga_feature_list *list );

struct fpga_feature_list *fpgafeat_init(struct device *dev);
void fpgafeat_deinit(struct fpga_feature_list **plist);

#endif // ETN_FPGA_MGR_FPGA_FEAT__H
