#define pr_fmt(fmt) "ETN_FPGA: %s:%d: " fmt, __FUNCTION__, __LINE__

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fpga.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/firmware.h>

#include "altera.h"
#include "main.h"
#include "fpgafeat.h"
#include "etn_fpga.h"

struct etn_fpga *etn_fpga = NULL;

int etn_fpga_state_get(void)
{
	int state = -1;
	struct fpga_manager *mgr = NULL; 

	if (!etn_fpga) {
		pr_err("etn_fpga is NULL\n");
		return -EINVAL;
	}

	mgr = platform_get_drvdata(etn_fpga->pdev);
	if (!mgr) {
		pr_err("Invalid mgr from container_of\n");
		return -EINVAL;
	}

	state = alt_fpga_state_get(mgr);
	if (state == ALT_FPGAMGR_STAT_USER_MODE) {
		return ETN_FPGA_STATE_USER_MODE;
	} else {
		return ETN_FPGA_STATE_INVALID;
	}
}
EXPORT_SYMBOL(etn_fpga_state_get);

int etn_fpga_ref_get(void)
{
	if (!etn_fpga) {
		pr_err("etn_fpga is NULL\n");
		return -EINVAL;
	}

	atomic_inc(&etn_fpga->refcnt);
	pr_debug("Inc refcnt = %d\n", atomic_read(&etn_fpga->refcnt));
	return 0;
}
EXPORT_SYMBOL(etn_fpga_ref_get);

int etn_fpga_ref_put(void)
{
	if (!etn_fpga)
		return -EINVAL;

	atomic_dec(&etn_fpga->refcnt);
	pr_debug("Dec refcnt = %d\n", atomic_read(&etn_fpga->refcnt));
	return 0;
}
EXPORT_SYMBOL(etn_fpga_ref_put);

struct fpga_feature *etn_fpga_get_feature(int type)
{
	if (etn_fpga && etn_fpga->features)
		return get_feature(type, etn_fpga->features);
	else
		return NULL;
}
EXPORT_SYMBOL(etn_fpga_get_feature);

uint16_t etn_fpga_feature_cr_base_on_port(struct fpga_feature *feat, 
                                          unsigned int port)
{
	return get_cr_base_on_port(feat, port);
}
EXPORT_SYMBOL(etn_fpga_feature_cr_base_on_port);

uint16_t etn_fpga_feature_sr_base_on_port(struct fpga_feature *feat, 
                                          unsigned int port)
{
	return get_sr_base_on_port(feat, port);
}
EXPORT_SYMBOL(etn_fpga_feature_sr_base_on_port);

#define ETN_FPGA_PATH "fpga/fpga.rbf"

/**
 * Flash FPGA firmware
 *
 * If we failed to flash FPGA we just move on in probe.
 *
 * @param fpga_path - path to fpga file relative to /lib/firmware
 * @param pdev      - pointer to platform device needed for resource accounting
 *                    and getting FPGA manager
 *
 *
 */
void etn_fpga_load_fw(char *fpga_path, struct platform_device *pdev)
{
	const struct firmware *fw = NULL;
	struct fpga_manager *mgr = NULL; 
	int rc = 0;	
	char *fw_data = NULL;

	rc = request_firmware(&fw, fpga_path, &pdev->dev);
	if (rc) {
		pr_err("Request FPGA firmware '%s' failed: %d\n", fpga_path, rc);
		return;
	}
	pr_info("Got firmware '%s', size %zu\n", fpga_path, fw->size);

	mgr = platform_get_drvdata(pdev);
	if (!mgr) {
		pr_err("Invalid FPGA mgr\n");
		rc = -EINVAL;
		goto release;
	}

	/* Copy fw from const pointer */
	fw_data = vmalloc(fw->size);
	if (!fw_data) {
		pr_err("Failed to allocate memory for firmware data\n");
		rc = -ENOMEM;
		goto release;
	}
	memcpy(fw_data, fw->data, fw->size);

	rc = alt_fpga_load_fw(fw_data, fw->size, mgr);
	if (rc) {
		pr_err("Failed to flash FPGA: %d\n", rc);
		goto release;
	}
	pr_info("Firmware loaded\n");

	if (etn_fpga_state_get() != ETN_FPGA_STATE_USER_MODE) {
		pr_err("FPGA is not configured\n");
		rc = -EIO;
		goto release;
	}
	pr_info("FPGA is in user mod\n");

release:
	vfree(fw_data);
	release_firmware(fw);
	return;
}

int etn_fpga_probe(struct platform_device *pdev)
{
	int err = 0;
	struct fpga_feature_list *flist;

	etn_fpga = kzalloc(sizeof(*etn_fpga), GFP_KERNEL);
	if (!etn_fpga)
		return -ENOMEM;

	etn_fpga->pdev = pdev;
	atomic_set(&etn_fpga->refcnt, 0);

	err = register_fpga_manager(pdev, &altera_fpga_mgr_ops,
	                            "ETN FPGA Manager", etn_fpga);
	if (err)
		goto err_free;

	if (etn_fpga_state_get() == ETN_FPGA_STATE_USER_MODE) {
		flist = fpgafeat_init(&pdev->dev);
		if (!flist) {
			pr_err("Failed to initialize FPGA features\n");
			goto err_unregister;
		}
		etn_fpga->features = flist;
	}

	/* Try to flash FPGA */
	etn_fpga_load_fw(ETN_FPGA_PATH, pdev);

	return 0;

err_unregister:
	remove_fpga_manager(pdev);
err_free:
	if (etn_fpga)
		kfree(etn_fpga);
	return -EINVAL;
}

int etn_fpga_remove(struct platform_device *pdev)
{
	remove_fpga_manager(pdev);
	fpgafeat_deinit(&etn_fpga->features);
	kfree(etn_fpga);
	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id etn_fpga_of_match[] = {
	{ .compatible = "etn,fpga-mgr", },
	{},
};

MODULE_DEVICE_TABLE(of, etn_fpga_of_match);
#endif

static struct platform_driver etn_fpga_driver = {
	.remove = etn_fpga_remove,
	.driver = {
		.name	= "etn_fpga_manager",
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(etn_fpga_of_match),
	},
};

static int __init etn_fpga_init(void)
{
	return platform_driver_probe(&etn_fpga_driver, etn_fpga_probe);
}

static void __exit etn_fpga_exit(void)
{
	platform_driver_unregister(&etn_fpga_driver);
}

module_init(etn_fpga_init);
module_exit(etn_fpga_exit);

MODULE_AUTHOR("Alan Tull <atull@altera.com>");
MODULE_AUTHOR("Alex Dzyoba <a.dzyoba@metrotek.spb.ru>");
MODULE_DESCRIPTION("ETN FPGA Manager");
MODULE_LICENSE("GPL v2");
MODULE_VERSION(DRV_VERSION);
