/**
 * File operations for /dev/etn char device.
 */
#define pr_fmt(fmt) "ETN_IO: %s:%d: " fmt, __FUNCTION__, __LINE__

#include <linux/fs.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "main.h"
#include "fops.h"
#include "fpga.h"

int etn_fpga_open(struct fpga_io *io)
{
	struct fpga_dev *etn = io->handle;

	if (!etn)
		return -EINVAL;

	pr_debug("Open! etn %p\n", etn);
	return 0;
}

/**
 * /dev/etn open callback (see etn_fops).
 *
 * Lock device and saves device descriptor in filp->private_data.
 */
static int etn_open(struct inode *inode, struct file *filp)
{
	int err = 0;
	struct fpga_dev *etn = 
		container_of(filp->private_data, struct fpga_dev, miscdev);
	struct fpga_io *io = etn->io;

	err = etn_fpga_open(io);
	if (err) {
		return err;
	}

	filp->private_data = etn;
	return 0;
}

int etn_fpga_close(struct fpga_io *io)
{
	struct fpga_dev *etn = io->handle;

	pr_debug("Release! etn %p, %p\n", etn, io);

	if (!etn) 
		return -EINVAL;
	return 0;
}

/**
 * /dev/etn device close callback (see etn_fops).
 */
static int etn_release(struct inode *inode, struct file *filp)
{
	struct fpga_dev *etn = filp->private_data;
	struct fpga_io *io = etn->io;
	return etn_fpga_close(io);
}

/**
 * /dev/etn seek callback (see etn_fops)
 *
 * Adjust filp->f_pos.
 * Does not respect SEEK_DATA, SEEK_HOLE, SEEK_MAX whence values.
 */
static loff_t etn_llseek(struct file *filp, loff_t off, int whence)
{
	loff_t newpos;

	pr_debug("llseek! off %lld, whence %d\n", off, whence);

	switch (whence) {
	case SEEK_SET:
		newpos = off;
		break;
	case SEEK_CUR:
		newpos = filp->f_pos + off;
		break;
	case SEEK_END:
		newpos = IOMAP_LEN + off;
		break;
	default:
		return -EINVAL;
	}

	if (newpos < 0 || newpos >= IOMAP_LEN) {
		return -EPROTO;
	}

	filp->f_pos = newpos;
	pr_debug("Returning newpos %lld\n", newpos);
	return newpos;
}

int etn_fpga_read(struct fpga_io* io, uint16_t reg, uint16_t *value)
{
	struct fpga_dev *etn = io->handle;
	*value = ioread16(etn->H2F + sizeof(*value) * reg);
	return 0;
}

/** 
 * /dev/etn read callback (see etn_fops)
 *
 * Do ioread8 (by byte) from mapped memory.
 */
ssize_t etn_read(struct file *filp, char __user *buf, 
                 size_t count, loff_t *f_pos)
{
	struct fpga_dev *etn = filp->private_data;
	size_t nleft, nxfer = 1;
	uint8_t temp;
	void __iomem *addr;

	pr_debug("Read: count %zu\n", count);

	if (count == 0 || *f_pos >= IOMAP_LEN)
		return -EPROTO;

	// Trim I/O
	if (*f_pos + count >= IOMAP_LEN)
		count = IOMAP_LEN - *f_pos;

	nleft = count;
	while (nleft > 0) {

		addr = etn->H2F + *f_pos;
		temp = ioread8(addr);
		pr_debug("  data[%lld] = 0x%x\n", *f_pos, temp);

		if (copy_to_user(buf, &temp, nxfer)) {
			return -EFAULT;
		}

		nleft  -= nxfer;
		buf    += nxfer;
		*f_pos += nxfer;
	}

	pr_debug("  Return %zd\n", count - nleft);
	return (count - nleft);
}

int etn_fpga_write(struct fpga_io *io, uint16_t reg, uint16_t value)
{
	struct fpga_dev *etn = io->handle;
	
	iowrite16(value, etn->H2F + sizeof(value) * reg);
	return 0;
}

/** 
 * /dev/etn write callback (see etn_fops)
 *
 * Do iowrite8 (by byte) to mapped memory.
 */
ssize_t etn_write(struct file *filp, const char __user *buf, 
                  size_t count, loff_t *f_pos)
{
	struct fpga_dev *etn = filp->private_data;
	size_t nleft, nxfer = 1;
	uint8_t temp;
	void __iomem *addr;

	pr_debug("Write: count %zu\n", count);

	if (count == 0 || *f_pos >= IOMAP_LEN) {
		pr_debug("  Return 0. count %zu, *f_pos %lld\n", count, *f_pos);
		return 0;
	}

	// Trim I/O
	if (*f_pos + count >= IOMAP_LEN)
		count = IOMAP_LEN - *f_pos;

	nleft = count;
	while (nleft > 0) {
		if (copy_from_user(&temp, buf, nxfer)) {
			return -EFAULT;
		}

		addr = etn->H2F + *f_pos;

		pr_debug("  data[%lld] = 0x%x\n", *f_pos, temp);
		iowrite8(temp, addr);

		nleft  -= nxfer;
		buf    += nxfer;
		*f_pos += nxfer;
	}

	pr_debug("  Return %zu\n", count - nleft);
	return (count - nleft);
}

const struct file_operations etn_fops = {
	.owner   = THIS_MODULE,
	.open    = etn_open,
	.release = etn_release,
	.llseek  = etn_llseek,
	.read    = etn_read,
	.write   = etn_write,
};

