/** 
 * ETN char device driver.
 *
 * Implements /dev/etn with open/close/read/write operations.
 * Provides I/O access to FPGA.
 * Actual I/O redirected to mapped FPGA memory.
 */

/*
 * TODO list:
 *
 * 1. Done. Move to platform driver
 * 2. Macro for request_mem_region/ioremap
 * 3. Configurable length for mapping instead of H2F_LEN. From dtb.
 * 4. Configuration parameters - choose interface to map. Nope.
 */

#define pr_fmt(fmt) "ETN_IO: %s:%d: " fmt, __FUNCTION__, __LINE__

#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/of.h>
#include <linux/dma-mapping.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>

#include <asm/uaccess.h>
#include <linux/ioport.h>
#include <asm/io.h>

#include <libfpgaio/fpgaio.h>
#include <etn_fpga_mgr/etn_fpga.h>

#include "main.h"
#include "fpga.h"
#include "fops.h"
#include "regs.h"



static struct fpga_dev *etn_fpga = NULL;

void __iomem *etn_io_regs_base(void)
{
	if (!etn_fpga)
		return NULL;
	else
		return etn_fpga->H2F;
}
EXPORT_SYMBOL(etn_io_regs_base);

/**
 *  Map FPGA memory for H2F and LWH2F interfaces.
 *  
 *  @param etn -  ETN device pointer
 *  @return 0 on success, -ENXIO on error.
 */
static int map_hw_mem(struct fpga_dev *etn)
{
	if (!request_mem_region(H2F_ADDR, H2F_LEN, DRV_NAME)) {
		pr_err("Failed to request mem region for H2F interface\n");
		goto err;
	}

	etn->H2F = ioremap(H2F_ADDR, H2F_LEN);
	if (!etn->H2F) {
		pr_err("Failed to map H2F address space\n");
		goto err_release_h2f;
	}

	if (!request_mem_region(LWH2F_ADDR, LWH2F_LEN, DRV_NAME)) {
		pr_err("Failed to request mem region for LWH2F interface\n");
		goto err_release_h2f;
	}

	etn->LWH2F = ioremap(LWH2F_ADDR, LWH2F_LEN);
	if (!etn->LWH2F) {
		pr_err("Failed to map LWH2F address space \n");
		goto err_release_lwh2f;
	}

	return 0;

err_release_lwh2f:
	if (etn->LWH2F) {
		iounmap(etn->LWH2F);
	}
	release_mem_region(LWH2F_ADDR, LWH2F_LEN);

err_release_h2f:
	if (etn->H2F) {
		iounmap(etn->H2F);
	}
	release_mem_region(H2F_ADDR, H2F_LEN);

err:
	return -ENXIO;
}

/**
 * Request memory mappings for H2F and LWH2F interfaces.
 */
static int init_hw_interface(struct fpga_dev *etn)
{
	if (map_hw_mem(etn)) {
		pr_err("Failed to map H2F and LWH2F memory\n");
		return -EIO;
	}
	return 0;
}

/**
 * Unmap memory requested for H2F and LWH2F interfaces
 */
static void unmap_hw_mem(struct fpga_dev *etn)
{
	if (etn->H2F) {
		iounmap(etn->H2F);
		etn->H2F = NULL;
		release_mem_region(H2F_ADDR, H2F_LEN);
	}

	if (etn->LWH2F) {
		iounmap(etn->LWH2F);
		etn->LWH2F = NULL;
		release_mem_region(LWH2F_ADDR, LWH2F_LEN);
	}
}

/**
 * Hardware interface deinitialization. Just unmaps memory
 */
static void deinit_hw_interface(struct fpga_dev *etn)
{
	unmap_hw_mem(etn);
}

extern const struct file_operations etn_fops;

static const struct miscdevice etn_miscdevice = {
	.minor = MISC_DYNAMIC_MINOR,
	.name  = DEV_NAME,
	.fops  = &etn_fops,
};

void etn_free(struct fpga_dev *fpga)
{
	if (fpga) {
		if( fpga->miscdev.this_device ) {
			misc_deregister(&fpga->miscdev);
			fpga->miscdev.this_device = NULL;
		}

		// Deinit hardware interfaces
		deinit_hw_interface(fpga);

		kfree(fpga);
	}
}

static struct fpga_io *fpga_kern_io_alloc(struct fpga_dev *fpga)
{
	struct fpga_io *io;
	io = kzalloc(sizeof(*io), GFP_KERNEL);
	if (!io) {
		pr_err("Failed to allocate FPGA I/O handler\n");
		return NULL;
	}

	io->open        = etn_fpga_open;
	io->close       = etn_fpga_close;
	io->read        = etn_fpga_read;
	io->write       = etn_fpga_write;
	io->handle      = fpga;
	io->cr_offset   = 81;
	io->sr_offset   = 82;
	return io;
}

static u64 platform_dma_mask = DMA_BIT_MASK(32);

static int etn_io_probe(struct platform_device *pdev)
{
	struct fpga_dev *fpga;
	struct fpga_io *io;

	if (etn_fpga_state_get() != ETN_FPGA_STATE_USER_MODE) {
		pr_err("Won't register I/O module when FPGA is not flashed\n");
		return -EIO;
	}

	if (etn_fpga_ref_get()) {
		pr_err("Failed to get ref on FPGA\n");
		return -EIO;
	}

	// fpga_dev is no longer allocated as netdev private, so do it manually
	fpga = kzalloc(sizeof(*fpga), GFP_KERNEL);
	if (!fpga) {
		pr_err("Failed to allocate memory for /dev/%s\n", DEV_NAME);
		goto err_put;
	}

	// Init hardware interfaces
	/* TODO: Must pass pdev so Initialize will get FPGA addresses from dtb */
	if (init_hw_interface(fpga)) {
		pr_err("Hardware interface configuration failed\n");
		goto err_free_fpgadev;
	}

	// ------ FPGA I/O handle -----
	io = fpga_kern_io_alloc(fpga);
	if (!io)
		goto err_free_fpgadev;

	fpga->io = io;

	pr_debug("Registering miscdev\n");
	fpga->miscdev = etn_miscdevice;
	if (misc_register(&fpga->miscdev)) {
		pr_err("Cannot register misc device\n");
		goto err_free_io;
	}

	pdev->dev.dma_mask = &platform_dma_mask;

	etn_fpga = fpga;
	fpga->pdev = pdev;
	platform_set_drvdata(pdev, fpga);

	return 0;

err_free_io:
	if (io)
		kfree(io);

err_free_fpgadev:
	if (fpga)
		kfree(fpga);

err_put:
	etn_fpga_ref_put();

	return -EINVAL;
}

static int etn_io_remove(struct platform_device *pdev)
{
	struct fpga_dev *fpga;

	fpga = platform_get_drvdata(pdev);
	if (fpga)
		etn_free(fpga);

	/* We don't need FPGA device anymore */
	etn_fpga_ref_put();
	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id etn_io_of_match[] = {
	{ .compatible = "etn,io", },
	{ /* sentinel */ },
};
MODULE_DEVICE_TABLE(of, etn_io_of_match);
#endif

static struct platform_driver etn_io_driver = {
	.remove = etn_io_remove,
	.driver = {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(etn_io_of_match),
	},
};

static int __init etn_io_init(void)
{
	/* XXX:
	 *   Don't try to use platform_driver_register here because 
	 *   it always return 0 even if probe failed, so the module 
	 *   will always load.  
	 */ 
	return platform_driver_probe(&etn_io_driver, etn_io_probe);
}

static void __exit etn_io_exit(void)
{
	platform_driver_unregister(&etn_io_driver);
}

MODULE_AUTHOR("Denis Gabidullin <d.gabidullin@metrotek.spb.ru>");
MODULE_AUTHOR("Alex Dzyoba <a.dzyoba@metrotek.spb.ru>");
MODULE_DESCRIPTION("Bercut ETN I/O access to Altera SoC FPGA interface");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);

module_init(etn_io_init);
module_exit(etn_io_exit);
