/** 
 * ETN network device driver.
 *
 * Uses etn_io driver for FPGA communication.
 */

#define pr_fmt(fmt) "ETN_NET: %s:%d: " fmt, __FUNCTION__, __LINE__

#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>

#include <linux/platform_device.h>
#include <linux/of_irq.h>
#include <linux/of_net.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/phy.h>

#include <etn_fpga_mgr/etn_fpga.h>
#include <etn_fpga_mgr/fpgafeat.h>
#include <etn_io/regs.h>

#include "main.h"
#include "net.h"
#include "mdio.h"



static int etn_init_fpga_regs(struct etn_netdev *etn_port)
{
	int port_num = etn_port->port_num;

	struct fpga_feature *nic, *phy;

	nic = etn_fpga_get_feature(FPGA_FEATURE_NIC_CTRL);
	phy = etn_fpga_get_feature(FPGA_FEATURE_MX_PHY);
	if (!nic || !phy)
		return -1;

	if (port_num == 0) {
		// DMA register intermixed, so we set 
		// base address to control register
		etn_port->dma.cr = MSGDMA_BASE_PORT0;
	} else if (port_num == 1) {
		etn_port->dma.cr = MSGDMA_BASE_PORT1;
	} else {
		return -1;
	}

	etn_port->nic.cr = etn_fpga_feature_cr_base_on_port(nic, port_num);
	etn_port->nic.sr = etn_fpga_feature_sr_base_on_port(nic, port_num);

	etn_port->phy.cr = etn_fpga_feature_cr_base_on_port(phy, port_num);
	etn_port->phy.sr = etn_fpga_feature_sr_base_on_port(phy, port_num);

	return 0;
}


/** 
 * Reset netdev and mdio bus of ETN port.
 */
static int __reset_netdev(struct etn_netdev *etn_port)
{
	int ret;

	pr_debug("Resetting port %p %d with netdev %p on pdev %p\n", 
	          etn_port, etn_port->port_num, etn_port->netdev, etn_port->pdev);

        etn_net_check_mac_address(etn_port);

	if (etn_init_fpga_regs(etn_port)) {
		pr_err("Can't initialize FPGA features for port %d\n", etn_port->port_num);
		goto err;
	}

	fpga_phy_reset(etn_port);

	pr_debug("Registering MDIO bus\n");
	ret = mdiobus_register(etn_port->mdio_bus);
	if (ret) {
		pr_err("Cannot register MDIO bus, err %d\n", ret);
		goto err;
	}

	pr_debug("Registering netdev\n");
	if (register_netdev(etn_port->netdev)) {
		pr_err("Cannot register net device for port %d\n", etn_port->port_num);
		goto err_mdio;
	}


	// Simple get first PHY device, because we have
	// only one PHY on each MDIO bus.
	etn_port->phy_dev = phy_find_first(etn_port->mdio_bus);
	if (!etn_port->phy_dev) {
		pr_debug("No PHY found\n");
		goto err_net;
	} 

	pr_debug("Connecting phy\n");
	ret = phy_connect_direct(etn_port->netdev, etn_port->phy_dev, 
	        &etn_handle_link_change, PHY_INTERFACE_MODE_RGMII_ID);

	if (ret) {
		pr_err("Could not attach to PHY\n");
		goto err_net;
	}

	pr_alert("PHY device '%s' for etn port %d\n", 
			etn_port->phy_dev->drv->name, etn_port->port_num);

	/* No half duplex */
	etn_port->phy_dev->supported = SUPPORTED_10baseT_Full |
	                               SUPPORTED_100baseT_Full |
	                               SUPPORTED_1000baseT_Full |
	                               SUPPORTED_Autoneg |
	                               SUPPORTED_TP | /* twisted pair */
	                               SUPPORTED_FIBRE;

	etn_port->phy_dev->advertising = etn_port->phy_dev->supported;

	// FIXME:
	//   May be we must turn off all LEDs for the first time
	if (etn_phy_conf_leds(etn_port, MII_88E1510_PHY_LED_COPPER)) {
		pr_err("Failed to initialize PHY LEDs\n");
		goto err_net;
	}

	return 0;

err_net:
	unregister_netdev(etn_port->netdev);
err_mdio:
	mdiobus_unregister(etn_port->mdio_bus);
err:
	return -EFAULT;
}

static void __net_unregister(struct etn_netdev *etn_port)
{
	if (etn_port) {
		pr_debug("Unregister port %d\n", etn_port->port_num);

		if (etn_port->mdio_bus) {
			if (etn_port->mdio_bus->state == MDIOBUS_REGISTERED)
				mdiobus_unregister(etn_port->mdio_bus);

			mdiobus_free(etn_port->mdio_bus);
			etn_port->mdio_bus = NULL;
		}

		if (etn_port->netdev) {
			if (netif_device_present(etn_port->netdev)) {
				cancel_work_sync(&etn_port->restart_work);
				unregister_netdev(etn_port->netdev);
			}
		}
	
		fpga_phy_off(etn_port);
	}
}

static int etn_register(struct etn_netdev *etn_port)
{
	if (etn_fpga_state_get() != ETN_FPGA_STATE_USER_MODE) {
		pr_err("FPGA is not ready\n");
		return -ENXIO;
	}

	if (__reset_netdev(etn_port)) {
		pr_err("Failed to reset ETN port %d\n", etn_port->port_num);
		goto err_net;
	}

	return 0;

err_net:
	__net_unregister(etn_port);
	return -EFAULT;
}

static void etn_unregister(struct etn_netdev *etn_port)
{
	if (etn_port) {
		pr_debug("Disconnecting phy\n");
		if (etn_port->phy_dev) {
			phy_disconnect(etn_port->phy_dev);
			etn_port->phy_dev = NULL;
	        }
	}
}

static void __netdev_free(struct etn_netdev *etn_port)
{
	if (etn_port) {
		__net_unregister(etn_port);
		free_netdev(etn_port->netdev);
	}
}

int etn_netdev_remove(struct platform_device *pdev)
{
	struct etn_netdev *etn_port;

	etn_port = platform_get_drvdata(pdev);
	if (etn_port) {
		pr_debug("Removing etn_port %p %d with netdev %p on pdev %p\n",
				etn_port, etn_port->port_num, etn_port->netdev, pdev);
		etn_unregister(etn_port);
		__netdev_free(etn_port);
	}
	etn_fpga_ref_put();
	platform_set_drvdata(pdev, NULL);
	return 0;      
}

static struct etn_netdev *etn_netdev_alloc(struct platform_device *pdev, int port_num)
{
	struct etn_netdev *etn_port = NULL;
	struct net_device *netdev = NULL;
	const u8 *mac_addr;

	netdev = alloc_etherdev(sizeof(struct etn_netdev));
	if (!netdev) {
		dev_err(&pdev->dev, "Cannot allocate netdev\n");
		goto err;
	}
	pr_debug("Allocated netdev %p on pdev %p\n", netdev, pdev);

	SET_NETDEV_DEV(netdev, &pdev->dev);

	etn_port = netdev_priv(netdev);
	etn_port->netdev = netdev;

	mac_addr = of_get_mac_address(pdev->dev.of_node);
	if (mac_addr)
		memcpy(netdev->dev_addr, mac_addr, ETH_ALEN);

	// MAC-address for eth0 is taken from .dts
	// MAC-address for eth1 is 'eth0+1' 
	if( port_num != 0 ) {
		if( netdev->dev_addr[5] == 0xFF ) {
			netdev->dev_addr[5] = 0;
			netdev->dev_addr[4]++;
		} else {
			netdev->dev_addr[5]++;
		}
	} 

	etn_port->port_num = port_num;

	etn_port->rx_data_size = netdev->mtu + ETH_HLEN + ETH_FCS_LEN;

	etn_port->regs_base = etn_io_regs_base();
	if (!etn_port->regs_base) {
		dev_err(&pdev->dev, "Can't get regsiters base address from etn_io\n");
		goto err_free_netdev;
	}

	// Current IRQ:
	//   72 (taken from .dtb)            -- Port0, TX
	//   73 (inc in etn_net.c:fpga_open) -- Port0, RX
	//   74 (taken from .dtb)            -- Port1, TX
	//   75 (inc in etn_net.c:fpga_open) -- Port1, RX
	netdev->irq = irq_of_parse_and_map(pdev->dev.of_node, 0) + 0;

	netdev->netdev_ops = &etn_netdev_ops;

	/* Explicitly turn off carrier to be able to flash FPGA */
	netif_carrier_off(netdev);

	netdev->ethtool_ops = &etn_ethtool_ops;

	netif_napi_add(netdev, &etn_port->napi, etn_net_poll, NAPI_WEIGHT);
	INIT_WORK(&etn_port->restart_work, etn_net_restart_work);

	etn_port->mdio_bus = etn_mdio_bus_alloc(etn_port, port_num);
	if (!etn_port->mdio_bus)
		goto err_free_netdev;

	return etn_port;

err_free_netdev:
	free_netdev(etn_port->netdev);
err:
	return NULL;
}

static u64 platform_dma_mask = DMA_BIT_MASK(32);

int etn_netdev_probe(struct platform_device *pdev)
{
	struct etn_netdev *etn_port = NULL;
	const __be32 *prop = NULL;
	int port_num = -1;

	if (etn_fpga_ref_get()) {
		pr_err("Failed to get ref on FPGA\n");
		return -EINVAL;
	}

	/* Get port number from dtb */
	prop = of_get_property(pdev->dev.of_node, "etn,port-num", NULL);
	if (!prop) {
		pr_err("No DeviceTree property for etn_net device");
		goto err_free_netdev;
	}
	port_num = be32_to_cpup(prop);

	etn_port = etn_netdev_alloc(pdev, port_num);
	if (!etn_port)
		goto err_free_netdev;

	pr_debug("Allocated etn_port %p %d on pdev %p\n", etn_port, port_num, pdev);
	
        pdev->dev.dma_mask = &platform_dma_mask;
	etn_port->pdev = pdev;
	platform_set_drvdata(pdev, etn_port);

	if (etn_register(etn_port)) {
		dev_err(&pdev->dev,"Cannot register devices\n");
		goto err_free_netdev;
	}

	return 0;

err_free_netdev:
	platform_set_drvdata(pdev, NULL);
	__netdev_free(etn_port);
	etn_fpga_ref_put();
	return -EINVAL;
}

#ifdef CONFIG_OF
static const struct of_device_id etn_of_match[] = {
	{ .compatible = "etn,net", },
	{ /* sentinel */ },
};
MODULE_DEVICE_TABLE(of, etn_of_match);
#endif

static struct platform_driver etn_net_driver = {
	.remove = etn_netdev_remove,
	.driver = {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(etn_of_match),
	},
};

static int __init etn_net_init(void)
{
	/* 
	 * Don't try to use platform_driver_register here because 
	 * it always return 0 even if probe failed, so the module 
	 * will always load.  
	 */ 
	return platform_driver_probe(&etn_net_driver, etn_netdev_probe);
}

static void __exit etn_net_exit(void)
{
	platform_driver_unregister(&etn_net_driver);
}

MODULE_AUTHOR("Denis Gabidullin <d.gabidullin@metrotek.spb.ru>");
MODULE_AUTHOR("Alex Dzyoba <a.dzyoba@metrotek.spb.ru>");
MODULE_DESCRIPTION("Bercut ETN driver fro FPGA network devices");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);

module_init(etn_net_init);
module_exit(etn_net_exit);
