#define pr_fmt(fmt) "ETN_NET: %s:%d: " fmt, __FUNCTION__, __LINE__

#include <linux/phy.h>
#include <linux/mii.h>

#include "main.h"
#include "regs.h"
#include "mdio.h"

/** 
 * Perform specific MDIO command
 * Support Clause 22 only.
 *
 * @param fpga - FPGA device pointer.
 * @param op - Code operation:
 *               MDIO_OP_WRITE -- write data.
 *               MDIO_OP_READ -- read data.
 * @param data - Data value. Ignored in MDIO_OP_READ.
 * @return error code (negative value) if error,
 *         0 if write operation success, 
 *         data if read operating success.
 *         TODO:
 *           Check work when error code returned.
 */
static int fpga_mdio_command(struct etn_netdev *port, int op, u16 data)
{
	if (fpga_fw_wait(port, port->phy.sr + MDIO_SR, MDIO_SR_BUSY, 0))
		return -ETIME;

	if (op == MDIO_OP_WRITE)
		fpga_fw_write_reg(port, port->phy.cr + MDIODATALO_CR, data);

	fpga_fw_write_pos(port, port->phy.cr + MDIO_CR,
			  MDIO_CR_COP_B1,
			  MDIO_CR_COP_B0,
			  op);

	fpga_fw_clear_bit(port, port->phy.cr + MDIO_CR, MDIO_CR_RUN);
	fpga_fw_set_bit(port, port->phy.cr + MDIO_CR, MDIO_CR_RUN);

	if (op == MDIO_OP_READ) {
		if (fpga_fw_wait(port, port->phy.sr + MDIO_SR, MDIO_SR_DATAVAL, 1))
			return -ETIME;

		return fpga_fw_read_reg(port, port->phy.sr + MDIODATALO_SR );
	}

	return 0;
}


/** 
 * Perform MDIO read command. 
 * Support Clause 22 only.
 *
 * @param bus - MDIO bus pointer.
 * @param mii_id - PHY address.
 * @param regnum - Register number.
 * @return data value if success, error code (negative) otherwise.
 */
static int fpga_mdio_read(struct mii_bus *bus, int mii_id, int regnum)
{
        struct etn_netdev *port = bus->priv;

        // Set PHY address 
        fpga_fw_write_pos(port, port->phy.cr + MDIOPHYAD_CR, 
                          MDIOPHYAD_CR_PHYAD_B4, MDIOPHYAD_CR_PHYAD_B0, mii_id);

        // Set register address
        fpga_fw_write_pos(port, port->phy.cr + MDIODEVAD_CR, 
                          MDIODEVAD_CR_DEVAD_B4, MDIODEVAD_CR_DEVAD_B0, regnum);


        // Read data
	return fpga_mdio_command(port, MDIO_OP_READ, 0);
}



/** 
 * Perform MDIO write command. 
 * Support Clause 22 only.
 *
 * @param bus - MDIO bus pointer.
 * @param mii_id - PHY address.
 * @param regnum - Register number.
 * @param data - Data for writing.
 * @return 0 if success, error code otherwise.
 */
static int fpga_mdio_write(struct mii_bus *bus, int mii_id, int regnum, u16 data)
{
        struct etn_netdev *port = bus->priv;
        
        // Set PHY address 
        fpga_fw_write_pos(port, port->phy.cr + MDIOPHYAD_CR, 
                          MDIOPHYAD_CR_PHYAD_B4, MDIOPHYAD_CR_PHYAD_B0, mii_id);

        // Set register address
        fpga_fw_write_pos(port, port->phy.cr + MDIODEVAD_CR, 
                          MDIODEVAD_CR_DEVAD_B4, MDIODEVAD_CR_DEVAD_B0, regnum);

        // Write data
	return fpga_mdio_command(port, MDIO_OP_WRITE, data);
}


void fpga_phy_reset(struct etn_netdev *port)
{
        fpga_phy_off(port);
        mdelay(1);
        fpga_phy_on(port);
        
        // XXX:
        //   Time for PHY turning on.
	mdelay(10); 
}

void fpga_phy_off(struct etn_netdev *port)
{
        fpga_fw_clear_bit(port, port->phy.cr + TRX_CR, TRX_CR_EG_NRST);
}

void fpga_phy_on(struct etn_netdev *port)
{
        fpga_fw_set_bit(port, port->phy.cr + TRX_CR, TRX_CR_EG_NRST);
}

/** 
 * Perform MDIO reset.
 *
 * @param bus - MDIO bus pointer.
 * @param mii_id - PHY address.
 * @param regnum - Register number.
 * @param data - Data for writing.
 * @return 0 if success, error code otherwise.
 */
static int fpga_mdio_reset(struct mii_bus *bus)
{
        struct etn_netdev *port = bus->priv;

        // Reset MDIO IP core 
        fpga_fw_set_bit(port, port->phy.cr + MDIO_CR, MDIO_CR_RST);

        // Set MDIO clock division factor 
        fpga_fw_write_pos(port, port->phy.cr + MDIODIV_CR, 
                          MDIODIV_CR_DIV_B5, MDIODIV_CR_DIV_B0, 50);

        // Return MDIO IP core to its normal operating mode 
        fpga_fw_clear_bit(port, port->phy.cr + MDIO_CR, MDIO_CR_RST);

        return 0;
}


/** 
 * Select speed mode in FPGA MAC core.
 * TODO:
 *   Now only one port is supported -- add second port.
 *
 * @param fpga - FPGA device pointer.
 * @param speed - Speed mode (10/100/1000). 
 * @param port_num - Port index. Unused now. 
 */
static void fpga_set_port_speed(struct etn_netdev *port, int speed, int port_num)
{
        int speed_for_fpga;
 
        switch(speed) {
        case 10:
                speed_for_fpga = PORT_SPEED_10;
                break;
        case 100:
                speed_for_fpga = PORT_SPEED_100;
                break;
        case 1000:
                speed_for_fpga = PORT_SPEED_1000;
                break;
        default:
                speed_for_fpga = PORT_SPEED_1000;
                pr_warn("Unsupported port speed. Set 1000 instead\n");
                break;

        }

        // TODO:
        //   Check, may we write this setting to FPGA realtime?
        fpga_fw_write_pos(port, port->phy.cr + PORT_SPEED_CR, 
                                PORT_SPEED_CR_RX_SPEED_B3, 
                                PORT_SPEED_CR_RX_SPEED_B0, 
                                speed_for_fpga);

        fpga_fw_write_pos(port, port->phy.cr + PORT_SPEED_CR, 
                                PORT_SPEED_CR_TX_SPEED_B3, 
                                PORT_SPEED_CR_TX_SPEED_B0, 
                                speed_for_fpga);
}

/** 
 * Apply current speed mode in FPGA MAC core .
 *
 * @param fpga - FPGA device pointer.
 */
static void fpga_apply_phy_settings(struct etn_netdev *port)
{
        fpga_set_port_speed(port, port->old_speed, 0);
}


/* 
 * Orig phy_speed_to_str() located in drivers/net/phy/phy.c
 * So copypaste, sorry.
 */
static const char *phy_speed_to_str(int speed)
{   
        switch (speed) {
        case SPEED_10:
                return "10Mbps";
        case SPEED_100:
                return "100Mbps";
        case SPEED_1000:
                return "1Gbps";
        case SPEED_2500:
                return "2.5Gbps";
        case SPEED_10000:
                return "10Gbps";
        case SPEED_UNKNOWN:
                return "Unknown";
        default:
                return "Unsupported";
        }
}   


/* 
 * Print interface status.
 * Differ from phy_print_status() -- also print port type (fiber/copper).
 */
void etn_phy_print_status(struct phy_device *phydev)
{
        if (phydev->link) {
                netdev_info(phydev->attached_dev,
                            "Link is Up (%s) - %s/%s - flow control %s\n",
                            phydev->port == PORT_FIBRE ? "Fiber" : "Copper",
                            phy_speed_to_str(phydev->speed),
                            DUPLEX_FULL == phydev->duplex ? "Full" : "Half",
                            phydev->pause ? "rx/tx" : "off");
        } else  {
                netdev_info(phydev->attached_dev, "Link is Down\n");
        }
}


/* 
 * Configure LED indication:
 */
int etn_phy_conf_leds(struct etn_netdev *etn_port, u16 leds_conf)
{
	int oldpage;
	struct mii_bus *bus = etn_port->mdio_bus;
	struct phy_device *phy = etn_port->phy_dev;

	if (!bus) {
		pr_err("Invalid MDIO bus pointer\n");
		return -EINVAL;
	}
	
	if (!phy) {
		pr_err("Invalid PHY device pointer\n");
		return -EINVAL;
	}

	/* Use low-level write callbacks because 
	 * phy_write/mdiobus_write lock bus mutex */
	mutex_lock(&bus->mdio_lock);
	oldpage = bus->read(bus, phy->addr, MII_MARVELL_PHY_PAGE);
	bus->write(bus, phy->addr, MII_MARVELL_PHY_PAGE, MII_88E1510_PHY_LED_PAGE);
	bus->write(bus, phy->addr, MII_88E1510_PHY_LED_CTRL, leds_conf);
	bus->write(bus, phy->addr, MII_MARVELL_PHY_PAGE, oldpage);
	mutex_unlock(&bus->mdio_lock);

	return 0;
}


/** 
 * Function called by phy_state_machine
 * when it will detect link/speed/duplex mode changes.
 *
 * For now we are looking only on link and speed mode.
 */
void etn_handle_link_change(struct net_device *netdev)
{
        struct etn_netdev *port = netdev_priv(netdev);
        struct phy_device *phy_dev = port->phy_dev;

        int status;
        int speed_changed = 0;
        int port_changed = 0;
        int status_changed = 0;
        u16 leds_conf;

        status = phy_read(phy_dev, MII_BMSR);
        if (status < 0) {
                pr_warn("Failed to read PHY status\n");
        }

        /* 
         * BUGFIX: When one of the port is in autoneg and other is not, 
         * phy_dev reports link as down, but netdev don't get it. 
         * So we set it here manually.
         * TODO: Maybe we can check here only PHY link status?
         */
        if ((status & BMSR_LSTATUS) == 0 && phy_dev->autoneg == 0)
                netif_carrier_off(netdev);

        pr_debug("%s Link change: speed %d, duplex %d, autoneg %d, link %d\n", netdev_name(netdev),
                        phy_dev->speed, phy_dev->duplex, phy_dev->autoneg, status & BMSR_LSTATUS);


        if (phy_dev->link) {
                if (port->old_speed != phy_dev->speed) {
                        port->old_speed = phy_dev->speed;
                        status_changed = 1;
                        speed_changed = 1;
                }

                if (port->old_port != phy_dev->port) {
                        port->old_port = phy_dev->port;
                        status_changed = 1;
                        port_changed = 1;
                }
        }

        if (phy_dev->link != port->old_link) {
                port->old_link = phy_dev->link;
                status_changed = 1;
        }


        if (status_changed) {
                if (speed_changed) {
                        fpga_apply_phy_settings(port);
                }

                if (port_changed) {
                        if (phy_dev->port == PORT_FIBRE) {
                                leds_conf = MII_88E1510_PHY_LED_FIBER;
                        } else {
                                leds_conf = MII_88E1510_PHY_LED_COPPER;
                        }

	                if (etn_phy_conf_leds(port, leds_conf)) {
		                pr_err("Failed to configuration PHY LEDs\n");
                        }
	        }

                etn_phy_print_status(phy_dev);
        }
}


struct mii_bus *etn_mdio_bus_alloc(struct etn_netdev *port, int port_num)
{
	int i;
	struct mii_bus *mdio;

	mdio = mdiobus_alloc();
	if (!mdio) {
		pr_err("Cannot allocate mdio_bus\n");
		return NULL;
	}

	mdio->name = "etn-mdio";
	mdio->read = &fpga_mdio_read;
	mdio->write = &fpga_mdio_write;
	mdio->reset = &fpga_mdio_reset;
	mdio->priv = port;

	// 0 -- check all (32) PHY addresses.
	// May be it is worth to decrease address count.
	mdio->phy_mask = 0;
	snprintf(mdio->id, MII_BUS_ID_SIZE, "etn-%d", port_num);

	mdio->irq = kmalloc(sizeof(int) * PHY_MAX_ADDR, GFP_KERNEL);
	if (!mdio->irq) {
		pr_err("Cannot allocate mdio->irq\n");
                mdiobus_free(mdio);
		return NULL;
	}

	// Select polling mode.
	for (i = 0; i < PHY_MAX_ADDR; i++)
		mdio->irq[i] = PHY_POLL;

	return mdio;
}


/**
 * Perform MDIO bus initialization.
 * FIXME:
 *  *) Create deinit function
 *  *) Select proper name for buses.
 *
 * TODO: 
 *  *) Change printk -> something else.
 *
 * @param fpga - FPGA device pointer.
 * @return 0 if success, error code otherwise.
 */
/*
 *int etn_mdio_bus_init(struct fpga_dev *fpga)
 *{
 *        int ret;
 *        struct mii_bus *mdio;
 *        struct phy_device *phydev;
 *
 *        mdio = etn_mdio_bus_alloc(fpga);
 *        if (!mdio) {
 *                printk("Cannot allocate mdio_bus\n");
 *                return -ENOMEM;
 *        }
 *
 *        fpga->mdio_bus = mdio;
 *
 *        ret = mdiobus_register(mdio);
 *        if (ret) {
 *                printk("Cannot register MDIO bus\n");
 *                goto err_free_mdio;
 *        }
 *
 *        // Simple get first PHY device, because we have
 *        // only one PHY on each MDIO bus.
 *        phydev = phy_find_first(mdio);
 *        if (!phydev) {
 *                printk("No PHY found\n");
 *                goto err_unregister_mdio;
 *        } 
 *
 *        // ID mean "internal delay".
 *        // We can enable/disable RX/TX delays separately
 *        // (PHY_INTERFACE_MODE_RGMII_RX_ID/PHY_INTERFACE_MODE_RGMII_TX_ID).
 *        // But for ETN we must turn on both delays, 
 *        // or it will cause timing violations.
 *        ret = phy_connect_direct(fpga->netdev, phydev, 
 *                                 &etn_handle_link_change, 
 *                                 PHY_INTERFACE_MODE_RGMII_ID);
 *        if (ret) {
 *                printk("Could not attach to PHY\n");
 *                goto err_unregister_mdio;
 *        }
 *
 *
 *        // We support only Full duplex.
 *        // TODO:
 *        //   Check SUPPORTED_Autoneg
 *        phydev->supported &= SUPPORTED_10baseT_Full |
 *                             SUPPORTED_100baseT_Full |
 *                             SUPPORTED_1000baseT_Full | 
 *                             SUPPORTED_Autoneg;
 *
 *        phydev->advertising = phydev->supported;
 *
 *        fpga->phy_dev = phydev;
 *
 *        return 0;
 *
 *err_unregister_mdio:
 *        mdiobus_unregister(mdio);
 *
 *err_free_mdio:
 *        mdiobus_free(mdio);
 *
 *        return ret;
 *}
 */

