#define pr_fmt(fmt) "ETN_NET: %s:%d: " fmt, __FUNCTION__, __LINE__

#include <linux/platform_device.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/of_irq.h>
#include <linux/interrupt.h>
#include <linux/phy.h>

#include <libfpgaio/fpgaio.h>

#include "main.h"
#include "mdio.h"
#include "regs.h"
#include "net.h"

static void fpga_set_mac_address(struct etn_netdev *etn_port)
{
	struct net_device *netdev = etn_port->netdev;
	u8 *dev_addr = netdev->dev_addr;
	int mac_cr = etn_port->nic.cr + HOST_MAC_W0_CR;

	netdev_dbg(etn_port->netdev, "Set MAC Address %p\n", dev_addr);
	fpga_fw_write_reg(etn_port, mac_cr + 2, (dev_addr[5] << 8) | dev_addr[4]);
	fpga_fw_write_reg(etn_port, mac_cr + 1, (dev_addr[3] << 8) | dev_addr[2]);
	fpga_fw_write_reg(etn_port, mac_cr + 0, (dev_addr[1] << 8) | dev_addr[0]);
}

void etn_net_check_mac_address(struct etn_netdev *etn_port)
{
	struct net_device *netdev = etn_port->netdev;
	u8 *dev_addr = netdev->dev_addr;
	int mac_cr = etn_port->nic.cr + HOST_MAC_W0_CR;

	if (is_valid_ether_addr(dev_addr)) {
		netdev_dbg(etn_port->netdev, "MAC Address is specified by configuration\n");
		fpga_set_mac_address(etn_port);
	} else {
		/* Try reading MAC address from FPGA. */
		uint16_t w0, w1, w2;

		w0 = fpga_fw_read_reg(etn_port, mac_cr + 0);
	        w1 = fpga_fw_read_reg(etn_port, mac_cr + 1);
		w2 = fpga_fw_read_reg(etn_port, mac_cr + 2);

		dev_addr[0] = w0 & 0xFF;
		dev_addr[1] = w0 >> 8;
		dev_addr[2] = w1 & 0xFF;
		dev_addr[3] = w1 >> 8;
		dev_addr[4] = w2 & 0xFF;
		dev_addr[5] = w2 >> 8;

		if (is_valid_ether_addr(dev_addr)) {
			/* FPGA values are valid, so use them. */
			netdev_dbg(etn_port->netdev, "MAC Address is read from FPGA: %pM\n", dev_addr);
		} else {
			/* FPGA values are invalid, generate random MAC. */
			random_ether_addr(dev_addr);
			netdev_dbg(etn_port->netdev, "MAC Address is set to random_ether_addr\n");
			fpga_set_mac_address(etn_port);
		}
	}
}


static void fpga_msgdma_reset(struct etn_netdev *port, struct alt_msgdma_csr __iomem *csr)
{
	unsigned long timeout;

	iowrite32(CSR_RESET_MASK, &csr->control);
	fpga_write_flush(port);
	udelay(20);

	timeout = round_jiffies(jiffies + HZ);
	while (ioread32(&csr->status) & CSR_RESET_STATE_MASK) {
		if (time_after(jiffies, timeout)) {
			pr_warn("MSGDMA reset bit never cleared\n");
			break;
		}
	}

	/* Clear status (the top 9 bits are used and are Read/Clear) */
	iowrite32(0x1FF, &csr->status);
	iowrite32(0, &csr->control);
	fpga_write_flush(port);
}

static void fpga_write_rx_descriptor(struct etn_netdev *port,
				     u32 write_address,
				     u32 transfer_length)
{
	struct alt_msgdma_csr __iomem *csr = port->regs_base + port->dma.cr + MSGDMA_RX_CSR;
	struct alt_msgdma_standard_descriptor __iomem *descriptor;

	/* Spin until there is room for another descriptor */
	while (ioread32(&csr->status) & CSR_DESCRIPTOR_BUFFER_FULL_MASK) {
		/* Just hope nothing goes wrong and full bit goes down */
	}

	descriptor = port->regs_base + port->dma.cr + MSGDMA_RX_DESC_FIFO;

	iowrite32(0, &descriptor->read_address);
	iowrite32(write_address, &descriptor->write_address);
	iowrite32(transfer_length, &descriptor->transfer_length);
	wmb();

	/* Commit all the descriptor information into the descriptor FIFO */
	iowrite32(DESCRIPTOR_CONTROL_GO_MASK |
		  DESCRIPTOR_CONTROL_END_ON_EOP_MASK |
		  DESCRIPTOR_CONTROL_TRANSFER_COMPLETE_IRQ_MASK |
		  DESCRIPTOR_CONTROL_EARLY_TERMINATION_IRQ_MASK,
		  &descriptor->control);

	fpga_write_flush(port);
}

static void fpga_construct_tx_descriptor(
	struct alt_msgdma_extended_descriptor *descriptor,
	u32 read_address,
	u32 transfer_length,
	u32 sequence_number,
	int generate_sop,
	int generate_eop,
	int transfer_complete_irq)
{
	memset(descriptor, 0, sizeof(*descriptor));

	descriptor->read_address = read_address;
	descriptor->transfer_length = transfer_length;
	descriptor->sequence_number = sequence_number;
	descriptor->control = DESCRIPTOR_CONTROL_GO_MASK;

	if (generate_sop)
		descriptor->control |= DESCRIPTOR_CONTROL_GENERATE_SOP_MASK;

	if (generate_eop)
		descriptor->control |= DESCRIPTOR_CONTROL_GENERATE_EOP_MASK;

	if (transfer_complete_irq)
		descriptor->control |= DESCRIPTOR_CONTROL_TRANSFER_COMPLETE_IRQ_MASK;
	else
		descriptor->control |= DESCRIPTOR_CONTROL_EARLY_DONE_ENABLE_MASK;
}

static void fpga_write_tx_descriptor(struct etn_netdev *port,
	                             struct alt_msgdma_extended_descriptor *descriptor)
{
	struct alt_msgdma_csr __iomem *csr = port->regs_base + port->dma.cr + MSGDMA_TX_CSR;
	struct alt_msgdma_extended_descriptor __iomem *target_descriptor;

	/* Spin until there is room for another descriptor */
	while (ioread32(&csr->status) & CSR_DESCRIPTOR_BUFFER_FULL_MASK) {
		/* Just hope nothing goes wrong and full bit goes down */
	}

	target_descriptor = port->regs_base + port->dma.cr + MSGDMA_TX_DESC_FIFO;

	iowrite32(descriptor->read_address, &target_descriptor->read_address);
	iowrite32(descriptor->write_address, &target_descriptor->write_address);
	iowrite32(descriptor->transfer_length, &target_descriptor->transfer_length);
	iowrite32(descriptor->sequence_number, &target_descriptor->sequence_number);
	iowrite32(descriptor->read_burst_count, &target_descriptor->read_burst_count);
	iowrite32(descriptor->write_burst_count, &target_descriptor->write_burst_count);
	iowrite32(descriptor->read_stride, &target_descriptor->read_stride);
	iowrite32(descriptor->write_stride, &target_descriptor->write_stride);
	iowrite32(descriptor->read_address_high, &target_descriptor->read_address_high);
	iowrite32(descriptor->write_address_high, &target_descriptor->write_address_high);
	wmb();

	/* Commit all the descriptor information into the descriptor FIFO */
	iowrite32(descriptor->control, &target_descriptor->control);

	fpga_write_flush(port);
}

static void fpga_read_rx_response(struct etn_netdev *port,
				  struct alt_msgdma_response *response)
{
	struct alt_msgdma_response __iomem *rx_response =
		port->regs_base + port->dma.cr + MSGDMA_RX_RESPONSE;

	response->actual_bytes_transferred =
		ioread32(&rx_response->actual_bytes_transferred);

	/* This read pops the response FIFO */
	response->errors = ioread32(&rx_response->errors);
}

static void fpga_clean_rx_ring(struct etn_netdev *etn_port)
{
	int i;

	for (i = 0; i < RX_DESC_FIFO_DEPTH; i++) {
		if (etn_port->rx_ring[i].skb) {
			dma_unmap_single(&etn_port->pdev->dev,
					 etn_port->rx_ring[i].dma_addr,
					 etn_port->rx_data_size,
					 DMA_FROM_DEVICE);

			dev_kfree_skb_any(etn_port->rx_ring[i].skb);

			etn_port->rx_ring[i].skb = NULL;
			etn_port->rx_ring[i].dma_addr = 0;
		}
	}

	etn_port->rx_head = 0;
	etn_port->rx_tail = 0;
}

static void fpga_clean_tx_ring(struct etn_netdev *etn_port)
{
	int i;

	for (i = 0; i < TX_DESC_FIFO_DEPTH; i++) {
		if (etn_port->tx_ring[i].skb) {
			dma_unmap_single(&etn_port->pdev->dev,
					 etn_port->tx_ring[i].dma_addr,
					 etn_port->tx_ring[i].skb->len,
					 DMA_TO_DEVICE);

			dev_kfree_skb_any(etn_port->tx_ring[i].skb);

			etn_port->tx_ring[i].skb = NULL;
			etn_port->tx_ring[i].dma_addr = 0;
		}
	}

	etn_port->tx_head = 0;
	etn_port->tx_tail = 0;
}

static int fpga_add_rx_buffer(struct etn_netdev *etn_port)
{
	struct net_device *netdev = etn_port->netdev;
	struct sk_buff *skb;
	struct fpga_rx_buff *buff;
	unsigned int next_head;

	next_head = (etn_port->rx_head + 1) % RX_DESC_FIFO_DEPTH;

	if (unlikely(next_head == etn_port->rx_tail)) 
		return -EBUSY;

#ifdef MSGDMA_RX_UNALIGNED_ACCESSES
	skb = netdev_alloc_skb_ip_align(netdev, etn_port->rx_data_size);
#else
	skb = netdev_alloc_skb(netdev, etn_port->rx_data_size);
#endif
	if (unlikely(skb == NULL)) {
		netif_err(etn_port, drv, netdev,
			  "Failed to allocate a socket buffer for RX\n");
		return -ENOMEM;
	}

	buff = &etn_port->rx_ring[etn_port->rx_head];
	buff->skb = skb;
	buff->dma_addr = dma_map_single(&(etn_port->pdev->dev), skb->data,
					etn_port->rx_data_size,
					DMA_FROM_DEVICE);

	if (dma_mapping_error(&etn_port->pdev->dev, buff->dma_addr)) {
		netif_err(etn_port, drv, netdev, "Failed to map RX buffer\n");
		dev_kfree_skb_any(skb);
		buff->skb = NULL;
		buff->dma_addr = 0;
		return -ENOMEM;
	}

	fpga_write_rx_descriptor(etn_port, PCIE_AVMM_TXS + buff->dma_addr,
				 etn_port->rx_data_size);

	etn_port->rx_head = next_head;

	return 0;
}

static int fpga_init_rx_ring(struct etn_netdev *etn_port)
{
	int i, err;
	struct alt_msgdma_csr __iomem *csr = etn_port->regs_base + 
	                                     etn_port->dma.cr + MSGDMA_RX_CSR;

	/* Fill up the RX descriptor FIFO */
	for (i = 0; i < RX_DESC_FIFO_DEPTH; i++) {
		err = fpga_add_rx_buffer(etn_port);
		if (err == -ENOMEM) {
			fpga_clean_rx_ring(etn_port);
			return err;
		}
	}

	/* Enable IRQs for RX */
	iowrite32(CSR_GLOBAL_INTERRUPT_MASK, &csr->control);
	fpga_write_flush(etn_port);

	return 0;
}

static void fpga_enable_tx_irq(struct etn_netdev *port)
{
	struct alt_msgdma_csr __iomem *csr = port->regs_base + port->dma.cr + MSGDMA_TX_CSR;

	iowrite32(CSR_GLOBAL_INTERRUPT_MASK, &csr->control);
	fpga_write_flush(port);
}

void etn_net_restart_work(struct work_struct *work)
{
	struct etn_netdev *etn_port =
		container_of(work, struct etn_netdev, restart_work);
	struct net_device *netdev = etn_port->netdev;

	netif_tx_lock(netdev);

	netif_wake_queue(netdev);

	netif_tx_unlock(netdev);
}

void inline tx_poll(struct etn_netdev *etn_port)
{
	struct net_device *netdev = etn_port->netdev;
	struct alt_msgdma_csr __iomem *csr = etn_port->regs_base + etn_port->dma.cr + MSGDMA_TX_CSR;
	unsigned int tx_tail;

	uint8_t free_desc_exist = 0;

        /* Check for completed TX DMA transfers */
	if (etn_port->tx_tail != etn_port->tx_head) {

		tx_tail = etn_port->tx_tail;
	        
 		while (tx_tail != ( ioread16(&csr->read_seq ) + 1 ) &&
		       tx_tail != etn_port->tx_head)
		{
			netdev->stats.tx_packets++;
			netdev->stats.tx_bytes += etn_port->tx_ring[tx_tail].skb->len;

		        dma_unmap_single(&etn_port->pdev->dev,
					 etn_port->tx_ring[tx_tail].dma_addr,
					 etn_port->tx_ring[tx_tail].skb->len,
					 DMA_TO_DEVICE);

			dev_kfree_skb_any(etn_port->tx_ring[tx_tail].skb);

			etn_port->tx_ring[tx_tail].skb = NULL;
			etn_port->tx_ring[tx_tail].dma_addr = 0;

			tx_tail = (tx_tail + 1) % TX_DESC_FIFO_DEPTH;
			etn_port->tx_tail = tx_tail;
		}

		smp_mb();


		// We must wake up TX netif part only if we have free descriptors in FPGA.
		free_desc_exist = ( ( ( etn_port->tx_head + 1) % TX_DESC_FIFO_DEPTH ) != etn_port->tx_tail );

		// With CONFIG_PREEMPT schedule_work()
		// cause BUG: schedule while atomic, so 
		// I've put restart_work function content 
		// here as is.
#ifdef CONFIG_PREEMPT
		netif_tx_lock(netdev);

		if (free_desc_exist) 
		  if (unlikely(netif_queue_stopped(netdev))) 
		    netif_wake_queue(netdev);


		netif_tx_unlock(netdev);
#else
		/* Can't restart safely under softirq */
		if (free_desc_exist) 
		  if (unlikely(netif_queue_stopped(netdev))) 
                    schedule_work(&etn_port->restart_work);
#endif
	}
}

int inline rx_poll(struct etn_netdev *etn_port, int budget)
{
	struct net_device *netdev = etn_port->netdev;
	struct alt_msgdma_csr __iomem *csr = etn_port->regs_base + 
	                                     etn_port->dma.cr + MSGDMA_RX_CSR;
	struct sk_buff *skb;
	struct alt_msgdma_response response;
	unsigned int rx_tail, rx_bytes;
	int howmany = 0;
        
	/* Read out all available results */
	while (!(ioread32(&csr->status)	& CSR_RESPONSE_BUFFER_EMPTY_MASK) &&
	       (howmany < budget))
	{
		fpga_read_rx_response(etn_port, &response);
		if (unlikely(response.error)) {
			netif_warn(etn_port, rx_err, netdev,
				   "RX error: 0x%02X\n", response.error);
			netdev->stats.rx_errors++;
		}

		rx_tail = etn_port->rx_tail;
		skb = etn_port->rx_ring[rx_tail].skb;

                dma_unmap_single(&etn_port->pdev->dev,
				 etn_port->rx_ring[rx_tail].dma_addr,
				 etn_port->rx_data_size,
                                 DMA_FROM_DEVICE);

		etn_port->rx_ring[rx_tail].skb = NULL;
		etn_port->rx_ring[rx_tail].dma_addr = 0;

		rx_bytes = response.actual_bytes_transferred;

		netdev->stats.rx_packets++;
		netdev->stats.rx_bytes += rx_bytes;

		skb_put(skb, rx_bytes);
		skb->protocol = eth_type_trans(skb, netdev);
		skb->ip_summed = CHECKSUM_UNNECESSARY;
		netif_receive_skb(skb);

		etn_port->rx_tail = (etn_port->rx_tail + 1) % RX_DESC_FIFO_DEPTH;

		if (unlikely(fpga_add_rx_buffer(etn_port))) {
			netif_err(etn_port, drv, netdev,
				  "Failed to add a new descriptor to RX ring\n");
			BUG();
		}

		howmany++;
	}

        return howmany;
}

int etn_net_poll(struct napi_struct *napi, int budget)
{
	struct etn_netdev *etn_port = container_of(napi, struct etn_netdev, napi);
	struct alt_msgdma_csr __iomem *csr;
        unsigned int rx_done;

        
	rx_done = rx_poll(etn_port, budget/2);
        tx_poll(etn_port);
        rx_done += rx_poll(etn_port, rx_done);
        tx_poll(etn_port);


	/* If all packets processed, turn on RX IRQs. */
	if (rx_done < budget) {
		napi_complete(napi);
		csr = etn_port->regs_base + etn_port->dma.cr + MSGDMA_RX_CSR;
		iowrite32(CSR_GLOBAL_INTERRUPT_MASK, &csr->control);
		fpga_write_flush(etn_port);
	}

	return rx_done;
}

static irqreturn_t fpga_isr(int irq, void *dev_id)
{
	struct etn_netdev *etn_port = dev_id;
	struct alt_msgdma_csr __iomem *csr = etn_port->regs_base + 
	                                     etn_port->dma.cr + MSGDMA_RX_CSR;

	/* Turn off RX IRQs and enable NAPI */
	if (likely(napi_schedule_prep(&etn_port->napi))) {
		iowrite32(0, &csr->control);
		fpga_write_flush(etn_port);
		__napi_schedule(&etn_port->napi);
	}

	/* Clear RX IRQ status */
	iowrite32(CSR_IRQ_SET_MASK, &csr->status);
	fpga_write_flush(etn_port);

	/* Clear TX IRQ status */
	csr = etn_port->regs_base + etn_port->dma.cr + MSGDMA_TX_CSR;
	iowrite32(CSR_IRQ_SET_MASK, &csr->status);
	fpga_write_flush(etn_port);

	return IRQ_HANDLED;
}

static struct net_device_stats *etn_net_get_stats(struct net_device *netdev)
{
	return &netdev->stats;
}

static netdev_tx_t etn_net_start_xmit(struct sk_buff *skb,
				   struct net_device *netdev)
{
	struct etn_netdev *etn_port = netdev_priv(netdev);
	struct fpga_tx_buff *buff;
	unsigned int head, next_head;

	head = etn_port->tx_head;
	next_head = (head + 1) % TX_DESC_FIFO_DEPTH;

	buff = &etn_port->tx_ring[head];
	buff->skb = skb;
	buff->dma_addr = dma_map_single(&etn_port->pdev->dev, skb->data,
					skb->len, DMA_TO_DEVICE);

	if (dma_mapping_error(&etn_port->pdev->dev, buff->dma_addr)) {
		netif_err(etn_port, drv, netdev, "Failed to map TX buffer\n");
		buff->skb = NULL;
		buff->dma_addr = 0;
		return NETDEV_TX_BUSY;
	}

	if (unlikely(((next_head + 1) % TX_DESC_FIFO_DEPTH) == etn_port->tx_tail)) {
		/* We are about to take the last descriptor */
		netif_stop_queue(netdev);
	}

	// We must generate IRQ for every TX descriptor.
	// Instead kernel will waiting for free sk_buff and stop Tx part.
	fpga_construct_tx_descriptor(&buff->desc, PCIE_AVMM_TXS + buff->dma_addr,
				     skb->len, head, 1, 1, 1);

	fpga_write_tx_descriptor(etn_port, &buff->desc);

	etn_port->tx_head = next_head;

	return NETDEV_TX_OK;
}

static int etn_net_change_mtu(struct net_device *netdev, int new_mtu)
{
	struct etn_netdev *etn_port = netdev_priv(netdev);
	int new_rx_data_size;

	if (netif_running(netdev))
		return -EBUSY;

	if (new_mtu < ETH_ZLEN || new_mtu > ETH_JUMBO_MTU)
		return -EINVAL;

	new_rx_data_size = new_mtu + ETH_HLEN + ETH_FCS_LEN;
	if (new_rx_data_size > SC_FIFO_WIDTH * SC_FIFO_DEPTH)
		return -EINVAL;

	netdev->mtu = new_mtu;
	etn_port->rx_data_size = new_rx_data_size;

	return 0;
}

static int etn_net_open(struct net_device *netdev)
{
	struct etn_netdev *etn_port = netdev_priv(netdev);
	int err;

	netdev_dbg(etn_port->netdev, "Opening netdev\n");

	netif_carrier_off(netdev);

	/* Initialize MSGDMA */
	fpga_msgdma_reset(etn_port, etn_port->regs_base + etn_port->dma.cr + MSGDMA_RX_CSR);
	fpga_msgdma_reset(etn_port, etn_port->regs_base + etn_port->dma.cr + MSGDMA_TX_CSR);

	etn_port->rx_head = 0;
	etn_port->rx_tail = 0;
	etn_port->tx_head = 0;
	etn_port->tx_tail = 0;

	/* Prepare RX ring to receive packets */
	err = fpga_init_rx_ring(etn_port);
	if (err) {
		netif_err(etn_port, ifup, netdev,
				"Failed to initialize RX DMA descriptor ring\n");
		return err;
	}

	/* Enable IRQs for TX */
	fpga_enable_tx_irq(etn_port);

	// Register MSGDMA Tx interrupt
	err = request_irq(netdev->irq, fpga_isr, IRQF_SHARED, DRV_NAME, etn_port);
	if (err) {
		netif_err(etn_port, ifup, netdev,
				"Cannot allocate interrupt %d\n", netdev->irq);
		goto err_clean_rx_ring;
	}
	
	// Register MSGDMA Rx interrupt
	err = request_irq(netdev->irq+1, fpga_isr, IRQF_SHARED, DRV_NAME, etn_port);
	if (err) {
		netif_err(etn_port, ifup, netdev,
				"Cannot allocate interrupt %d\n", netdev->irq);
		goto err_clean_rx_ring;
	}

	fpga_set_mac_address(etn_port);

	/* Enable NAPI polling */
	napi_enable(&etn_port->napi);

	// In case of FPGA reflashing we must force writing to FPGA 
	// in etn_handle_link_change().
	etn_port->old_speed = -1;

	if(etn_port->phy_dev) {
		phy_start(etn_port->phy_dev);
	}

	/* Start transmit queue */
	netif_start_queue(netdev);

	return 0;

err_clean_rx_ring:
	fpga_clean_rx_ring(etn_port);
	return err;
}

static int etn_net_stop(struct net_device *netdev)
{
	struct etn_netdev *etn_port = netdev_priv(netdev);

	netdev_dbg(etn_port->netdev, "Stopping netdev %s\n", netdev->name);

	netif_carrier_off(netdev);
	netif_tx_disable(netdev);

	// Wait here for poll to complete
	napi_disable(&etn_port->napi); 

	if (etn_port->phy_dev)
		phy_stop(etn_port->phy_dev);

	fpga_msgdma_reset(etn_port, etn_port->regs_base + etn_port->dma.cr + MSGDMA_RX_CSR);
	fpga_msgdma_reset(etn_port, etn_port->regs_base + etn_port->dma.cr + MSGDMA_TX_CSR);

	// Free Rx and Tx IRQ
	free_irq(netdev->irq, etn_port);
	free_irq(netdev->irq+1, etn_port);

	fpga_clean_rx_ring(etn_port);
	fpga_clean_tx_ring(etn_port);

	return 0;
}

static int etn_ioctl(struct net_device *netdev, struct ifreq *rq, int cmd)
{
	struct etn_netdev *etn_port = netdev_priv(netdev);
	int rc = -EOPNOTSUPP;

	switch (cmd) {
		case SIOCGMIIPHY:
		case SIOCGMIIREG:
		case SIOCSMIIREG:
			if (!etn_port->phy_dev)
				return -EINVAL;

			rc = phy_mii_ioctl(etn_port->phy_dev, rq, cmd);
			break;
		default:
			break;
	}
	return rc;
}

const struct net_device_ops etn_netdev_ops = {
	.ndo_open	     = etn_net_open,
	.ndo_stop	     = etn_net_stop,
	.ndo_get_stats	     = etn_net_get_stats,
	.ndo_start_xmit      = etn_net_start_xmit,
	.ndo_change_mtu      = etn_net_change_mtu,
	.ndo_validate_addr   = eth_validate_addr,
	.ndo_set_mac_address = eth_mac_addr,
	.ndo_do_ioctl        = etn_ioctl,
};

static void etn_ethtool_get_drvinfo(struct net_device *netdev,
				     struct ethtool_drvinfo *info)
{
	u16 fw_version = 0;

	strlcpy(info->driver, DRV_NAME, sizeof(info->driver));
	strlcpy(info->version, DRV_VERSION, sizeof(info->version));

	snprintf(info->fw_version, sizeof(info->fw_version), "%u.%u.%u-%u",
		 (fw_version >> 12) & 0xF, (fw_version >> 8) & 0xF,
		 (fw_version >>  4) & 0xF, (fw_version >> 0) & 0xF);
}

static u32 etn_ethtool_get_msglevel(struct net_device *netdev)
{
	struct etn_netdev *etn_port = netdev_priv(netdev);
	return etn_port->msg_enable;
}

static void etn_ethtool_set_msglevel(struct net_device *netdev, uint32_t data)
{
	struct etn_netdev *etn_port = netdev_priv(netdev);
	etn_port->msg_enable = data;
}

static int etn_ethtool_get_settings(struct net_device *netdev, struct ethtool_cmd *cmd)
{
	int ret = 0;
	struct etn_netdev *etn_port = netdev_priv(netdev);
	if (etn_port == NULL || etn_port->phy_dev == NULL)
		return -ENODEV;

	ret = phy_ethtool_gset(etn_port->phy_dev, cmd);
	cmd->port = etn_port->phy_dev->port;
	return ret;
}

static int etn_ethtool_set_settings(struct net_device *netdev, struct ethtool_cmd *cmd)
{
	struct etn_netdev *etn_port = netdev_priv(netdev);
	if (etn_port == NULL || etn_port->phy_dev == NULL)
		return -ENODEV;

	return phy_ethtool_sset(etn_port->phy_dev, cmd);
}


const struct ethtool_ops etn_ethtool_ops = {
  .get_drvinfo	= etn_ethtool_get_drvinfo,
  .get_msglevel = etn_ethtool_get_msglevel,
  .set_msglevel = etn_ethtool_set_msglevel,
  .get_settings = etn_ethtool_get_settings,
  .set_settings = etn_ethtool_set_settings,
  .get_link     = ethtool_op_get_link,
};

