
#include <linux/platform_device.h>
#include "main.h"

// Flush function. Used only in PCIe driver. 
inline void fpga_write_flush(struct etn_netdev *port)
{
        return;
}

/** 
 * Read FPGA register.
 *
 * @param fpga - FPGA device pointer.
 * @param reg - Register number.
 * @return Readed data.
*/ 
u16 fpga_fw_read_reg(struct etn_netdev *port, int reg)
{
	u16 tmp;
	tmp = ioread16(port->regs_base + 2 * reg);
	return tmp;
}


/** 
 * Write FPGA register.
 *
 * @param fpga - FPGA device pointer.
 * @param reg - Register number. 
 * @param val - Data for write.
*/ 
void fpga_fw_write_reg(struct etn_netdev *port, int reg, u16 val)
{
	iowrite16(val, port->regs_base + 2 * reg);
	fpga_write_flush(port);
}


/** 
 * Read particular bits from FPGA register.
 *
 * @param fpga - FPGA device pointer.
 * @param reg - Register number. 
 * @param msb - Most significant bit for reading (included)
 * @param lsb - Least significant bit for reading (included)
 * @return Readed data.
*/ 
u16 fpga_fw_read_pos(struct etn_netdev *port,
                            int reg, int msb, int lsb)
{
	u16 mask = (1 << (msb - lsb + 1)) - 1, tmp;

	tmp = fpga_fw_read_reg(port, reg);
	return (tmp >> lsb) & mask;
}


/** 
 * Write particular bits to FPGA register.
 *
 * @param fpga - FPGA device pointer.
 * @param reg - Register number. 
 * @param msb - Most significant bit for writing (included).
 * @param lsb - Least significant bit for writing  (included).
 * @param val - Data for write.
*/ 
void fpga_fw_write_pos(struct etn_netdev *port,
                              int reg, int msb, int lsb, u16 val)
{
	u16 mask = (1 << (msb - lsb + 1)) - 1, tmp;

	tmp = fpga_fw_read_reg(port, reg) & ~(mask << lsb);
	fpga_fw_write_reg(port, reg, tmp | ((val & mask) << lsb));
}


/** 
 * Write '1' to selected bit in FPGA register.
 *
 * @param fpga - FPGA device pointer.
 * @param reg - Register number. 
 * @param bit - Bit index. 
*/ 
void fpga_fw_set_bit(struct etn_netdev *port, int reg, int bit)
{
	unsigned long tmp = fpga_fw_read_reg(port, reg);

	set_bit(bit, &tmp);
	fpga_fw_write_reg(port, reg, tmp);
}


/** 
 * Write '0' to selected bit in FPGA register.
 *
 * @param fpga - FPGA device pointer.
 * @param reg - Register number. 
 * @param bit - Bit index. 
*/ 
void fpga_fw_clear_bit(struct etn_netdev *port, int reg, int bit)
{
	unsigned long tmp = fpga_fw_read_reg(port, reg);

	clear_bit(bit, &tmp);
	fpga_fw_write_reg(port, reg, tmp);
}


/** 
 * Test selected bit in FPGA register.
 *
 * @param fpga - FPGA device pointer.
 * @param reg - Register number. 
 * @param bit - Bit index. 
 * @return 1 if bit set, 0 otherwise.
*/ 
int fpga_fw_test_bit(struct etn_netdev *port, int reg, int bit)
{
	unsigned long tmp = fpga_fw_read_reg(port, reg);

	return test_bit(bit, &tmp);
}


/** 
 * Wait until selected bit in FPGA register take required value.
 * Timeout 1 HZ.
 *
 * @param fpga - FPGA device pointer.
 * @param reg - Register number. 
 * @param bit - Bit index. 
 * @param sense - Required value.
 * @return 0 if success, -ETIME if timeout.
*/ 
int fpga_fw_wait(struct etn_netdev *port, int reg, int bit, int sense)
{
	unsigned long timeout = round_jiffies(jiffies + HZ);

	while (!!fpga_fw_test_bit(port, reg, bit) ^ !!sense) {
		if (time_after(jiffies, timeout)) {
			dev_warn(&port->pdev->dev,
				 "%s: bit never set or cleared: "
				 "reg=%d, bit=%d, sense=%d\n",
				 __func__, reg, bit, !!sense);
			return -ETIME;
		}
	}

	return 0;
}
