#ifndef __KERNEL__
#  include <stdio.h>
#  include <errno.h>
#else 
#  include <asm-generic/errno.h>
#endif

#include "fpgaio.h"

#ifdef __KERNEL__
  #include <linux/printk.h>
  #define ERR(format, ...) pr_err( "Error: "format"\n", ##__VA_ARGS__)
#else
  #define ERR(format, ...) fprintf( stderr, "Error: "format"\n", ##__VA_ARGS__)
#endif

/**
 *
 */
int fpga_open( struct fpga_io * io ) 
{
  if( !io ) {
    ERR( "%s: IO pointer is NULL\n", __func__ );
    return EINVAL;
  }

  if( !io->open ) {
    ERR( "%s: IO->open pointer is NULL\n", __func__ );
    return EINVAL;
  }

  return io->open( io );
}

/**
 *
 */
int fpga_close( struct fpga_io * io ) 
{
  if( !io ) {
    ERR( "%s: IO pointer is NULL\n", __func__ );
    return EINVAL;
  }

  if( !io->close ) {
    ERR( "%s: IO->close pointer is NULL\n", __func__ );
    return EINVAL;
  }

  return io->close( io );
}

/**
 *
 */
int fpga_read_reg( struct fpga_io *io, uint16_t reg, uint16_t *data )
{
  if( !io ) {
    ERR( "%s: IO pointer is NULL\n", __func__ );
    return EINVAL;
  }

  if( !io->read ) {
    ERR( "%s: IO->read pointer is NULL\n", __func__ );
    return EINVAL;
  }

  return io->read( io, reg, data );
}

/**
 *
 */
int fpga_write_reg( struct fpga_io *io, uint16_t reg, uint16_t data )
{
  
 if( !io ) {
    ERR( "%s: IO pointer is NULL\n", __func__ );
    return EINVAL;
  }

  if( !io->write ) {
    ERR( "%s: IO->write pointer is NULL\n", __func__ );
    return EINVAL;
  }

  return io->write( io, reg, data );
}

/**
 *
 */
int fpga_clear_bit( struct fpga_io *io, uint16_t reg, uint8_t bit )
{
  uint16_t tmp = 0;
  int err;

  if ((err = fpga_read_reg(io, reg, &tmp))) {
    return err;
  }

  tmp &= ~(1 << bit);

  return fpga_write_reg(io, reg, tmp);
}

/**
 *
 */
int fpga_set_bit( struct fpga_io *io, uint16_t reg, uint8_t bit )
{
  uint16_t tmp = 0;
  int err;

  if ( ( err = fpga_read_reg( io, reg, &tmp ) ) ) {
    return err;
  }

  tmp |= (1 << bit);

  return fpga_write_reg( io, reg, tmp );
}

/**
 *
 */
int fpga_fall_bit( struct fpga_io *io, uint16_t reg, uint8_t bit)
{
  int err = 0;
  if ( ( err = fpga_set_bit( io, reg, bit ) ) ) {
    return err;
  }
  return fpga_clear_bit( io, reg, bit );
}

/**
 *
 */
int fpga_rise_bit( struct fpga_io *io, uint16_t reg, uint8_t bit )
{
  // may be we need to check error code? or may be not..
  int err = 0;
  if ( ( err = fpga_clear_bit( io, reg, bit ) ) ) {
    return err;
  }
  return fpga_set_bit( io, reg, bit );
}

/**
 *
 */
int fpga_read_mask( struct fpga_io *io, uint16_t reg, uint16_t *data, 
                                        uint8_t msb, uint8_t lsb )
{
  int err = 0;
  uint16_t tmp = 0;
  if ( ( err = fpga_read_reg( io, reg, &tmp ) ) ) {
    return err;
  }

  (*data) = (tmp >> lsb) & ((1 << (msb - lsb + 1)) - 1);

  return 0;
}

/**
 *
 */
int fpga_write_mask( struct fpga_io *io, uint16_t reg, uint16_t data, 
                                         uint8_t msb, uint8_t lsb )
{
  uint16_t tmp = 0;
  int err = 0;

  uint16_t mask = ( 1 << ( msb - lsb + 1 ) ) - 1;

  if ( ( err = fpga_read_reg( io, reg, &tmp ) ) ) {
    return err;
  }


  // clear'em all
  tmp &= ~( mask << lsb );

  tmp |= ( ( data & mask ) << lsb );

  return fpga_write_reg( io, reg, tmp );
}

/**
 *
 */
int fpga_wait_for( struct fpga_io *io, uint16_t reg, uint8_t bit, 
                                       uint8_t level )
{
  uint16_t tmp = 0;
  int err = 0;

  while (1) {
    if ((err = fpga_read_reg(io, reg, &tmp)))
      return err;

    if (((tmp & (1 << bit)) ^ level) == 0)
      break;
  }

  return 0;
}
