/** \file io.h 
 *  \brief Fpga input/output wrapper functions.
 *         They wrap interface of fpga_io into interface
 *         which is used in auto-generated functions.
 */

#ifndef FPGA_IO__H
#define FPGA_IO__H

#ifndef __KERNEL__
#include <stdint.h>
#else
#include <linux/types.h>
#endif

/**
 * \brief Struct with all necessary information to communicate with FPGA 
 *        (for one FPGA feature).
 */
typedef struct fpga_io {
  /// Callback for low-level open functon
  int (*open)( struct fpga_io* io );

  /// Callback for low-level write functon
  int (*write)( struct fpga_io* io, uint16_t addr, uint16_t value );

  /// Callback for low-level read functon
  int (*read)( struct fpga_io* io, uint16_t addr, uint16_t *value );
  
  /// Callback for low-level close functon
  int (*close)( struct fpga_io* io );

  /// Handle, that used for access to fpga 
  void *handle;

  /// Feature CR offset
  uint16_t cr_offset;

  /// Feature SR offset
  uint16_t sr_offset;
  
} fpga_io_t;

/** \brief Open fpga device for IO operarion
 *  \param io Pointer to fpga_io struct
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_open( struct fpga_io *io );

/** \brief Close fpga device for IO operarion
 *  \param io Pointer to fpga_io struct
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_close( struct fpga_io *io );

/** \brief Read single FPGA register
 *  \param io Pointer to io_struct
 *  \param reg Address of register
 *  \param data Pointer to data
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_read_reg( struct fpga_io *io, uint16_t reg, uint16_t *data );

/** \brief Write data to single FPGA register
 *  \param io Pointer to io_struct
 *  \param reg Address of register
 *  \param data Data to write
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_write_reg( struct fpga_io *io, uint16_t reg, uint16_t data );

/** \brief Clear one bit in an FPGA register.
 *  \param io Pointer to io_struct
 *  \param reg Address of register
 *  \param bit Index of bit to clear
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_clear_bit( struct fpga_io *io, uint16_t reg, uint8_t bit );

/** \brief Set one bit in an FPGA register.
 *  \param io Pointer to io_struct
 *  \param reg Address of register
 *  \param bit Index of bit to set
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_set_bit( struct fpga_io *io, uint16_t reg, uint8_t bit );

/** \brief Set one bit in an FPGA register to 1 and then to 0 ( 1->0 )
 *  \param io Pointer to io_struct
 *  \param reg Address of register
 *  \param bit Index of bit to process
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_fall_bit(struct fpga_io *io, uint16_t reg, uint8_t bit );

/** \brief Set one bit in an FPGA register to 0 and then to 1 ( 0->1 )
 *  \param io Pointer to io_struct
 *  \param reg Address of register
 *  \param bit Index of bit to process
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_rise_bit( struct fpga_io *io, uint16_t reg, uint8_t bit );

/** \brief Read some masked value from an FPGA register
 *  \param io Pointer to io_struct
 *  \param reg Address of register
 *  \param data Pointer to data. Read data is shifted on lsb value, so it always
 *         places in least bits
 *  \param msb Mask most significant bit index
 *  \param lsb Mask least significant bit index
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_read_mask( struct fpga_io *io, 
                    uint16_t reg, uint16_t *data, 
                    uint8_t msb, uint8_t lsb );

/** \brief Write some masked value to an FPGA register
 *  \param io Pointer to io_struct
 *  \param reg Address of register
 *  \param data Unshifted data to write 
 *  \param msb Mask most significant bit index
 *  \param lsb Mask least significant bit index
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_write_mask( struct fpga_io *io, 
                     uint16_t reg, uint16_t data,
                     uint8_t msb, uint8_t lsb);

/** \brief Wait until FPGA register bit will reach specified level
 *  \param io Pointer to io_struct
 *  \param reg Address of register
 *  \param bit index of bit to wait 
 *  \param level Bit value that you expect
 *  \return 0 if there are no errors, error code otherwise
 */
int fpga_wait_for( struct fpga_io *io, 
                   uint16_t reg, uint8_t bit,
                   uint8_t level );

#endif /* ETN_IO__H */
